# Copyright 2007 Edd Dawson.
# Distributed under the Boost Software License, Version 1.0.
# See accompanying file BOOST_LICENSE_1_0.txt or copy at http://www.boost.org/LICENSE_1_0.txt

"""
This module contains the functionality that brings togther a macroscopic build description
anda toolset in order to create tasks to execute in order to carry out the build.
"""

import os
from stat import ST_MTIME

from slamcore.engine.taskpool import task
from slamcore.engine.taskpool import taskpool
from slamcore.engine import deps
import slamcore.engine.message as message

def remove_slam_directories(path):
	"""
	This function removes path and all of it's parent directories. It stops on the 
	directory that either does not contain a file named ".slam_made_this", or that
	is otherwise non-empty.
	"""
	slam_flag = '.slam_made_this'

	entries = []
	if os.path.exists(path) and os.path.isdir(path):
		entries = os.listdir(path)
	else:
		popped_path = os.path.split(path)[0]
		if popped_path and len(popped_path) > 0:
			remove_slam_directories(popped_path)

	if len(entries) == 1 and entries[0] == slam_flag:
		try:
			os.remove(os.path.join(path, slam_flag))
			message.info('removing directory ' + path)
			os.rmdir(path)
			popped_path = os.path.split(path)[0]
			if popped_path and len(popped_path) > 0:
				remove_slam_directories(popped_path)
		except:
			return

def generate_clean_queue_impl_(deps_db, proj, toolset, clean_finals):
	"""
	This function is used to implement generate_clean_queue and generate_tidy_queue. It returns a
	list of slamcore.engine.taskpool.taskpool objects that should be called in sequence to clean the build
	products. If clean_finals is False, the final target for each slamfile (e.g. the library or 
	executable) will NOT be removed.

	When processing the returned taskpool objects, you should finish processing each taskpool before
	moving on to the next. However, the tasks within a given taskpool can be executed in parallel.

	deps_db should be the slamcore.engine.deps.database object, proj the
	slamcore.engine.description.project object and toolset the slamcore.toolsets.toolset object for 
	the build.
	"""
	queues = []
	for prof in proj.get_profiles():
		q = taskpool()
		sources = prof.args('sources') or []
		for s in sources:
			q.add_queue(toolset.clean_tasks(s, prof))

		queues.append(q) # delete objects

		if clean_finals:
			q = taskpool()
			q.add_queue(toolset.clean_product_tasks(prof))
			queues.append(q) # delete target

		d = prof.args('dir')
		if d:
			q = taskpool()
			del_dirs = lambda d_=d[0]: remove_slam_directories(d_)
			del_dirs = task(del_dirs, 'removing ' + d[0] + ' and parent directories')
			q.add_queue([del_dirs])
			queues.append(q) # delete target directory (and intermediates)

	return queues
	
def generate_clean_queue(deps_db, proj, toolset):
	"""
	Calls generate_clean_queue_impl_(deps_db, proj, toolset, True).
	See the descrition of generate_clean_queue_impl_ for more information.
	"""
	return generate_clean_queue_impl_(deps_db, proj, toolset, True)
	
def generate_tidy_queue(deps_db, proj, toolset):
	"""
	Calls generate_clean_queue_impl_(deps_db, proj, toolset, False).
	See the descrition of generate_clean_queue_impl_ for more information.
	"""
	return generate_clean_queue_impl_(deps_db, proj, toolset, False)

def src_needs_building(deps_db, prof, toolset, src_file):
	"""
	Returns True if the source file with the given filename needs to be built.

	This function is slightly conservative; it returns True a little more often than it should.
	To make this function entirely correct, we would need to check that the commands to build this
	source file haven't changed from the last time that it was built. This information would have
	to be stored in deps_db.

	deps_db should be the slamcore.engine.deps.database object, and toolset the
	slamcore.toolsets.toolset object for the build. prof should be the
	slamcore.engine.description.profile that is currently being processed.
	"""
	ts_name = toolset.name()
	outfile = toolset.build_product(src_file, prof)

	if not outfile or not os.path.exists(outfile):
		return True

	outfile_time = os.stat(outfile)[ST_MTIME]
	src_file_time = os.stat(src_file)[ST_MTIME]

	if outfile_time < src_file_time:
		return True

	# Output file is older than last dependency calculation for src_file
	# Strictly speaking, this check is conservative
	if outfile_time < deps_db.get_update_time(prof, ts_name, src_file):
		return True

	deps = deps_db.get_dependencies(prof, ts_name, src_file)
	if deps:
		dep_files = deps.deps()
		for f in dep_files:
			if os.stat(f)[ST_MTIME] > outfile_time: return True
		
	return False

def target_needs_building(prof, prof_deps, updated_profs, toolset):
	"""
	Returns True if the final target of the slamcore.engine.description.profile given by prof needs
	to be built. prof_deps should be the list of profiles on which prof depends. updated_profs 
	should be a list of profiles (that is a subset of prof_deps) that will be updated as part of 
	the build. toolset should be the slamcore.toolsets.toolset object for the build
	"""
	target = toolset.target_product(prof)
	if not target: return False
	
	if not os.path.exists(target): return True	
	
	target_time = os.stat(target)[ST_MTIME]
	
	for dprof in prof_deps:
		if dprof in updated_profs:
			return True	
			
		dtarget = toolset.target_product(dprof)
		if dtarget and os.path.exists(dtarget) and os.stat(dtarget)[ST_MTIME] > target_time:
			return True

	objects = [toolset.build_product(s, prof) for s in prof.args('sources') or []]
	for obj in objects:
		if os.path.exists(obj) and os.stat(obj)[ST_MTIME] > target_time: 
			return True
		
	return False
	

def generate_build_queue(deps_db, proj, toolset):
	"""
	This function returns a list of slamcore.engine.taskpool.taskpool objects that should be called in 
	sequence to perform the build.

	When processing the returned taskpool objects, you should finish processing each taskpool before 
	moving on to the next. However, the tasks within a given taskpool can be executed in parallel.

	deps_db should be the slamcore.engine.deps.database object, proj the
	slamcore.engine.description.project object and toolset the slamcore.toolsets.toolset object for 
	the build.
	"""

	# This could possibly be improved:
	# Can we archive all libraries in parallel? 
	# Might have difficulties with some dynamic libs? -- investigate
	# Then we can link executables in parallel
	queues = []
	updated_profs = []
	queues.append(taskpool())

	for prof in proj.get_profiles():
		outfiles = []
		updates = False
		sources = prof.args('sources') or []
		prof_deps = proj.get_profile_dependencies(prof)
		
		for s in sources:
			if src_needs_building(deps_db, prof, toolset, s):
				queues[0].add_queue(toolset.build_tasks(s, prof))
				updates = True
			outfiles.append(toolset.build_product(s, prof))

		if updates:
			updated_profs.append(prof)
				
		if updates or target_needs_building(prof, prof_deps, updated_profs, toolset):
			q = taskpool()
			q.add_queue(toolset.target_tasks(prof, prof_deps))
			queues.append(q)

	return queues