# Copyright 2007 Edd Dawson.
# Distributed under the Boost Software License, Version 1.0.
# See accompanying file BOOST_LICENSE_1_0.txt or copy at http://www.boost.org/LICENSE_1_0.txt

"""
The txt.formatter class is the main export from this module. If formats blocks of text from toolset
documentation for output to the console.
"""

import string

class block_handler:
	"""
	This class is used by formatter in order to process a block of text. If the block appears to be
	a list of items demarkated by a '*' or a '-', special action is taken. Else the block is left
	alone.
	
	When a block is being processed, each line should be added to a block_handler by calling its
	add_line() method. When all lines of a block have been added, call the text() method to get back
	the formatted string.
	"""
	def __init__(self, icons = '*-'):
		self.lines_ = []
		self.icon_ = None
		self.icons_ = icons
		self.txt_ = ''

	def add_line(self, line):
		for i in self.icons_:
			if line.startswith(i + ' '): 
				if self.icon_ == i: self.lines_ += [line]
				else:
					# Changing icon, treat it as a new list
					# Flush the previous one
					self.flush_lines()
					self.icon_ = i
					self.lines_ = [line]
				break
		else:
			self.flush_lines()
			self.txt_ += '    ' + line + '\n'


	def flush_lines(self):
		if len(self.lines_) > 0:
			for l in self.lines_:
				if l[0] != '*': l = l[2:]  # 2: to remove icon and space
				self.txt_ += '    ' + l + '\n'
		self.lines_ = []
		
	def text(self):
		self.flush_lines()
		return self.txt_



class formatter:
	def preamble(self):
		"""
		This function returns a string containing a newline character.

		The formatter contract specifies that the preamble() method of a formatter implementation
		must return a string that should be put at the start of a document.
		"""
		return '\n'

	def postamble(self):
		"""
		This function returns an empty string.

		The formatter contract specifies that the postamble() method of a formatter implementation
		must return a string that should be put at the start of a document.
		"""
		return ''

	def title(self, t):
		"""
		Returns some text that contains t with an underline of '-' characters.

		The formatter contract specifies that the title() method of a formatter implementation
		should return a string that contains t formatted appropriately for display as a main title.
		"""
		return string.capitalize(t) + '\n' + (len(t) * '-') + '\n\n'

	def heading(self, g):
		"""
		Returns some text that contains g with ':' suffix.

		The formatter contract specifies that the heading() method of a formatter implementation
		should return a string that contains g formatted appropriately for display as a heading.
		"""
		return string.capitalize(g) + ':\n'

	def text(self, txt):
		"""
		Uses a block_handler object to format txt appropriately as a single string, which is 
		returned.
		
		The formatter contract specifies that the text() method of a formatter implementation
		should return a string that contains txt formatted appropriately for display as text which
		contributes to the body of a document (e.g. paragraphs, lists, etc).
		"""
		# split in to blocks, delimited by two or more \n characters
		lines = [l.strip() for l in txt.split('\n')]

		blocks = []
		block = []
		for l in lines:
			if len(l) == 0: # empty line
				if len(block):
					blocks.append(block)
				block = []
			else:
				block.append(l)

		if len(block): blocks.append(block)

		ret = ''
		for b in blocks:
			bh = block_handler()
			for line in b:
				bh.add_line(line)
			ret += bh.text() + '\n'

		return ret