# Copyright 2007 Edd Dawson.
# Distributed under the Boost Software License, Version 1.0.
# See accompanying file BOOST_LICENSE_1_0.txt or copy at http://www.boost.org/LICENSE_1_0.txt

"""
This module contains the facilities needed to manage source file dependency information
"""

import os
import pickle
import time
import threading
import math
from stat import ST_MTIME

import slamcore.engine.message as message
import slamcore.engine.taskpool as taskpool


def mod_time(filename):
	"""
	utility function to return the modification time of a file
	"""
	return os.stat(filename)[ST_MTIME]

def string_list_hash(strlist):
	"""
	Returns a "hash" for a list of strings. This has limited utility outside of this module.
	"""
	strlist2 = strlist[:]
	strlist2.sort()
	ret = ''
	for s in strlist2: ret += str(len(s)) + '|' + s
	return ret


class src_deps:
	"""
	Each instance of this class hold information about the dependencies of a particular source 
	file, including the time that the dependency information was last checked and/or modified,
	and an identifier for the facility/command that updated those dependencies.
	"""

	def __init__(self, update_cmd, dps):
		"""
		Constructs a src_deps object that record the dependencies in the list dps that were 
		generated by the command identified by the string in update_cmd.
		"""
		self.check_time_ = 0
		self.change_time_ = 0
		self.updater_id_ = ''
		self.deps_ = []
		self.update(update_cmd, dps)

	def last_checked(self):
		"""
		Returns the time that the recorded dependencies were last checked, whether they had changed 
		or not. The time is expressed in seconds since the epoch (UTC).
		"""
		return self.check_time_

	def last_changed(self):
		"""
		Returns the time that the recorded dependencies were last changed. The time is expressed in 
		seconds since the epoch (UTC).
		"""
		return self.change_time_

	def deps(self):
		"""
		Returns a list of strings that contains filenames of dependencies.
		"""
		return self.deps_

	def updater_id(self):
		"""
		Returns a string that represents the facility used to generate the dependency information 
		recorded in this object
		"""
		return self.updater_id_

	def update(self, updater_id, dps):
		"""
		If updater_id != self.updater_id() or dps != self.deps(), the new set of dependencies and
		the string representation for updating command is stored internally. Calling this function
		will affect the subsequent return value of last_checked(), and possibly last_changed()
		"""
		#if os.name == 'nt': self.check_time_ = math.floor(time.time())
		#else: self.check_time_ = time.time()
		self.check_time_ = math.floor(time.time())
		if dps != self.deps_ or updater_id != self.updater_id_:
			self.updater_id_ = updater_id
			self.deps_ = dps
			self.change_time_ = self.check_time_


class database:
	"""
	A database object holds all the dependency information for all source files that are involved
	in a build and coordinates the update of that information.

	It also deduces the order in which the sub-projects (aka slamcore.engine.description.profiles)
	should be built.
	"""
	database_version_ = 2

	def __init__(self, filename):
		"""
		Attempts to load a database from a given file. If the file cannot be found, the database
		construction succeeds but it will not contain any dependency information to begin with.
		"""
		self.abs_name_ = os.path.abspath(filename)
		self.db_ = {}
		self.mutex_ = threading.Lock()

		if os.path.exists(self.abs_name_) and os.path.isfile(self.abs_name_):
			f = open(self.abs_name_, 'rb')
			try:
				data = pickle.load(f)
				f.close()

				if len(data) != 3 or data[1] != database.database_version_:
					message.warn('slam.db data format has changed, dependencies will be recalculated')
				else:
					if os.path.realpath(data[0]) == os.path.realpath(self.abs_name_): self.db_ = data[2]
					else: message.warn('slam.db has moved, dependencies will be recalculated')
			except:
				message.warn('slam.db corrupted, dependencies will be recalculated')
				f.close()

	def save(self, name = None):
		"""
		Saves (pickles) the database to the file with the given filename, unless it is None, in which
		case the database is saved to the file with the name given at construction time, or the name
		previously passed to save().
		"""
		if name: self.abs_name_ = os.path.abspath(os.path.join(get_dir_part(sys.argv[0]), name))
		f = open(self.abs_name_, 'wb')
		pickle.dump((self.abs_name_, database.database_version_, self.db_), f)
		f.close()

	def out_of_date_(self, src, prof, updater_id, toolset_name):
		"""
		Returns True if src needs to be built. prof is the slamcore.engine.description.profile
		object that represents slamfile that controls the compilation of src. updater_id is the 
		string identifier for the command or facility that can be used to update the dependency
		information of src. toolset_name is the name of the toolset that may be used to compile
		the source file.
		"""
		cfg = prof.get_path()
		variants = string_list_hash(prof.get_variants())
		deps_data = self.db_[cfg][variants][toolset_name][src]

		last_checked = deps_data.last_checked()

		if mod_time(cfg) > last_checked: return True
		if mod_time(src) > last_checked: return True
		if updater_id != deps_data.updater_id(): return True

		for dep in deps_data.deps():
			if not os.path.exists(dep) or mod_time(dep) > last_checked:
				return True

		return False

	def per_thread_update_(self, src, prof, deps_dict, toolset):
		"""
		This function calculates the dependencies for src and stores them in deps_dict if they have
		changed. deps_dict is a dictionary that maps filename strings to src_deps objects.

		prof is the slamcore.engine.description.profile	object that represents slamfile that
		controls the compilation of src. toolset is the slamcore.toolsets.toolset.toolset object 
		that is being used to (potentially) build src.
		
		This function can be called multiple times in parallel to update the same deps_dict.
		"""
		tsk = toolset.deps(src, prof)
		if not deps_dict.has_key(src):
			message.info('updating dependencies for ' + src)
			deps = tsk()
			self.mutex_.acquire()
			deps_dict[src] = src_deps(tsk.description(), deps)
			self.mutex_.release()

		elif self.out_of_date_(src, prof, tsk.description(), toolset.name()):
			message.info('updating dependencies for ' + src)
			deps = tsk()
			self.mutex_.acquire()
			deps_dict[src].update(tsk.description(), deps)
			self.mutex_.release()

	def update(self, proj, toolset, num_threads):
		"""
		Checks and updates the dependencies for all the source files in the build described by
		proj, which is a slamcore.engine.description.project object. toolset is the 
		slamcore.toolsets.toolset.toolset object that can be used to compile and update the source
		files. num_threads is the number of source-file dependency calculations that may be carried
		out in parallel. Typically this should be set to the number of processors in the machine
		performing the build.
		"""
		queue = taskpool.taskpool()

		for prof in proj.get_profiles():
			sources = prof.args('sources')
			if not sources: continue

			cfg = prof.get_path()
			variants = string_list_hash(prof.get_variants())
			ts_name = toolset.name()

			if not self.db_.has_key(cfg): self.db_[cfg] = {}
			if not self.db_[cfg].has_key(variants): self.db_[cfg][variants] = {}
			if not self.db_[cfg][variants].has_key(ts_name): self.db_[cfg][variants][ts_name] = {}

			# Create a queue of commands that generate dependency information
			deps_dict = self.db_[cfg][variants][ts_name]
			for src in sources:
				cmd = lambda s=src, p=prof, d=deps_dict: self.per_thread_update_(s, p, d, toolset)
				seq = [taskpool.task(cmd, 'updating dependencies for ' + src)]
				queue.add_queue(seq)

		queue.execute(num_threads, False)

	def wipe_deps(self, proj, toolset):
		"""
		Clears all stored dependency information from the database relating to builds of the given
		slamcore.engine.description.project (proj) and slamcore.toolsets.toolset.toolset (toolset).
		"""
		ts_name = toolset.name()

		for prof in proj.get_profiles():
			variants = string_list_hash(prof.get_variants())
			cfg = prof.get_path()

			if self.db_.has_key(cfg) and self.db_[cfg].has_key(variants) and self.db_[cfg][variants].has_key(ts_name):
				del self.db_[cfg][variants][ts_name]


	def get_dependencies(self, prof, toolset_name, src_file):
		"""
		Returns the list of filenames stored in the database that src_file depends upon. 
		
		The dependency information is identified by the given toolset-name and
		slamcore.engine.description.profile object.
		"""
		cfg = prof.get_path()
		variants = string_list_hash(prof.get_variants())
		if 	self.db_.has_key(cfg):
			db = self.db_[cfg] 
			if db.has_key(variants): 
				db = self.db_[cfg][variants]
				if db.has_key(toolset_name):
					db = self.db_[cfg][variants][toolset_name]
					if db.has_key(src_file):
						return self.db_[cfg][variants][toolset_name][src_file]
		return None
					
	def get_update_time(self, prof, toolset_name, src_file):
		"""
		Returns the time in seconds since the epoch (UTC) that the dependency information in the
		database for src_file was changed, or 0 if the database has no dependency information about
		src_file.
		
		The dependency information is identified by the given toolset-name and
		slamcore.engine.description.profile object.
		"""
		deps_data = self.get_dependencies(prof, toolset_name, src_file)
		if deps_data:
			return deps_data.last_changed()
		else: 
			return 0
