# Copyright 2007 Edd Dawson.
# Distributed under the Boost Software License, Version 1.0.
# See accompanying file BOOST_LICENSE_1_0.txt or copy at http://www.boost.org/LICENSE_1_0.txt

"""
This module contains the facilities needed to descibe a build. The most important classes in here
are profile and project
"""

import os
import glob
import copy
import re

import slamcore.engine.parser as parser
import slamcore.engine.message as message
from slamcore.engine.env import expand_variables
from slamcore.toolsets import pkgconfig
from slamcore.engine.exceptions import slam_error


class slamfile:
	"""
	Instances of this class contain the parsed data from a particular slamfile
	"""

	def __init__(self, path):
		"""
		Reads the slamfile whose filename is given by path. An exception of type derived from
		slamcore.engine.exceptions.slam_error is thrown on failure
		"""
		self.data_ = {}
		self.path_ = os.path.normpath(path)

		parsed_lines = parser.parse_slamfile(path)
		for parts in parsed_lines:
			if not parts[0] in self.data_.keys(): self.data_[parts[0]] = {}
			if not parts[1] in self.data_[parts[0]].keys(): self.data_[parts[0]][parts[1]] = []
			self.data_[parts[0]][parts[1]] += parts[2]

	def get_path(self):
		"""
		Returns the path of the slamfile that was parsed to created this object
		"""
		return self.path_

	def args(self, directive, variant):
		"""
		Given the name of a directive and a build variant, a list of directive arguments will
		be returned. If the directive was not present in the slamfile for the given variant,
		None is returned. No replacements are made from the slam environment, here. $(XYZ)s
		will appear in the returned arguments for example.
		"""
		if not directive in self.data_: return None
		if not variant in self.data_[directive]: return None
		return self.data_[directive][variant]

	def get_data(self):
		"""
		Returns the dict of dicts that contains the parsed slamfile data. This dict of dicts maps
		directive-name -> dict of variants -> list of arguments
		"""
		return self.data_
		
		
class profile:
	"""
	A profile object is a snapshot of the data parsed from a slamfile for a particular variant
	and with variables expanded from the environment.
	"""
	def __init__(self, slamfile, env, variants, toolset):
		"""
		Constructs a profile object for the given slamfile object. All variables found in directive
		arguments are expanded from the slamcore.engine.env.env object given by env. 
		
		All the variants given in variants list are assumed relevant. Typically variants will be
		['common', proper_variant], but it could be anything else.
		
		All filename globbing is expanded here, too.
		
		If the slamcore.engine.toolsets.toolset.toolset object given by toolset supports pkg-config,
		the include dirs, defines and flags for the compiler and linker are added to the appropriate
		arguments.
		"""
		self.data_ = {}
		self.path_ = slamfile.get_path()
		self.variants_ = variants
		desc_data = slamfile.get_data()
		for directive in desc_data.keys():
			for variant in variants:
				values = slamfile.args(directive, variant)
				if values:
					if not self.data_.has_key(directive): self.data_[directive] = []
					self.data_[directive] += values

		self.expand_env_variables_(env)
		self.expand_sources_()
		self.expand_dirs_()
		self.integrate_pkg_config_(toolset)

	def args(self, directive):
		"""
		Returns the arguments for the given directive as a list, or None if the directive wasn't
		present in the given slamfile.
		"""
		if directive in self.data_.keys(): return self.data_[directive][:]
		return None

	def append_arg(self, directive, a):
		"""
		Appends a (string) to the arguments for the given directive
		"""
		if directive in self.data_.keys(): self.data_[directive].append(a)
		else: self.data_[directive] = [a]

	def get_variants(self):
		"""
		Returns the variants passed in on construction
		"""
		return self.variants_

	def get_path(self):
		"""
		Returns the location of the slamfile that corresponds to this profile.
		"""
		return self.path_;

	def get_dir(self):
		"""
		Returns the directory that contains the slamfile corresponding to this profile.
		"""
		return os.path.dirname(self.path_)
		
	def expand_env_variables_(self, env):
		"""
		Does the expansion of all the variables found in arguments. 
		
		Used by the constructor, shouldn't be called elsewhere.
		"""
		try:
			env_with_dir = copy.deepcopy(env)
			env_with_dir.set_variable('here', self.get_dir()) # use $(here) for the profile directory
			for k in self.data_.keys():
				values = self.data_[k]
				for i in range(0, len(values)):
					values[i] = expand_variables(values[i], env_with_dir)
				values = filter(len, values)

		except RuntimeError, var:
			raise slam_error("no match for '" + str(var) + "' in environment. Needed in " + self.path_)

	def expand_sources_(self):
		"""
		Does the globbing of all the source files in the arguments to the 'sources' directive(s). 

		Used by the constructor, shouldn't be called elsewhere.
		"""
		sources = self.args('sources')
		if sources:
			expanded = []
			d = os.path.dirname(self.path_)
			for s in sources:
				if os.path.isabs(s): expanded += glob.glob(s) # will this always work?
				else: expanded += glob.glob(os.path.join(d, s))
				
			self.data_['sources'] = expanded
			for s in expanded:
				if not (os.path.exists(s) and os.path.isfile(s)):
					raise slam_error("file '" + s + "' not found. Listed in " + self.path_)
			
	def expand_dirs_(self):
		"""
		Searches in the arguments of various directives that are known to be directory names and
		'normalises' the strings of all directories found.
		
		Used by the constructor, shouldn't be called elsewhere.
		"""
		for directive in ['include', 'dir', 'libdirs']:
			values = self.args(directive)
		
			if values:
				expanded = []
				for v in values:
					if os.path.isabs(v): expanded.append(v)
					else: expanded.append(os.path.normpath(os.path.join(self.get_dir(), v)))
				self.data_[directive] = expanded
		
			
	def integrate_pkg_config_(self, toolset):
		"""
		Adds arguments to some of the directives if there are any arguments to the pkgconfig 
		directive.
		
		Used by the constructor, shouldn't be called elsewhere.
		"""
		packages = self.args('pkg-config') or []
		if len(packages) == 0: return
		
		names = toolset.names()
		if 'msvc' not in names and 'gcc' not in names:
			message.warn('The pkgconfig directive probably only makes sense for msvc and gcc based toolsets')

		(defs, incs, ldirs, libs, cflags, lflags) = pkgconfig.parse(packages, 'msvc' in names)

		for d in defs: self.append_arg('defines', d)
		for i in incs: self.append_arg('include', i)
		for d in ldirs: self.append_arg('libdirs', d)
		for l in libs: self.append_arg('libs', l)
		if len(cflags) > 0:
			self.append_arg(toolset.name(), 'compiler: ' + ' '.join(cflags))
		if len(lflags) > 0:
			self.append_arg(toolset.name(), 'linker: ' + ' '.join(lflags))
			

def find_graph_path_(graph, start, end, route=[]):
	"""
	Guido's simple graph code from:
	http://www.python.org/doc/essays/graphs/
	"""
	route = route + [start]
	if start == end: return route
	if not graph.has_key(start): return None
	for node in graph[start]:
		if node.get_path() not in route:
			newroute = find_graph_path_(graph, node.get_path, end, route)
			if newroute: return newroute
	return None

def check_for_cycles_(graph):
	"""
	Returns a list of strings that form a cycle in the graph, if one exists. Else None is
	returned.
	"""
	for k in graph.keys():
		for p in graph[k]:
			route = find_graph_path_(graph, p.get_path(), k)
			
			if route:
				route = [k] + route
				return [r for r in route]
			
	return None # no cycles found


class project:
	"""
	A project object is essentially a collection of profile objects that describes the entire 
	build.
	"""
	config_name_ = 'slamfile'

	def __init__(self, env, variants, toolset, recurse):
		"""
		Contructs a project object that consists of build instructions from a number of slamfiles.
		
		If recursive is False, the slamfile located at the 'path' environment variable is sought.
		If recursive is True, all directories 'under' the location in the 'path' environment 
		variable are searched for slamfiles.

		In either case, the current directory is used if there is no 'path' variable in env.
		
		In either case, the 'depends' directives of the slamfiles are followed to find yet more 
		slamfiles relevant to the build.
		"""
		self.profiles_ = []
		self.variants_ = variants
		self.toolset_ = toolset
		self.dependencies_map_ = {} # maps a profile path on to profile objects
		self.ignore_dir_patterns_ = []

		self.fill_ignore_dir_patterns_()

		path = env.get_value('path') or ''

		if recurse: self.find_configurations_(path, env, variants, toolset)
		else: self.load_profiles_(path, env, variants, toolset, [])
		
		route = check_for_cycles_(self.dependencies_map_)
		if route:
			err = 'dependencies cycle found: ' + reduce(lambda x, y: x + ' -> ' + y, route)
			raise slam_error(err)

		def profile_compare_(proj, p1, p2):
			if p1 == p2: return 0
			if p1 in proj.dependencies_map_[p2.get_path()]: return -1
			return 1

		comp = lambda p1, p2: profile_compare_(self, p1, p2)
		self.profiles_.sort(comp)
		
	def fill_ignore_dir_patterns_(self):
		"""
		This function is called by the constructor to read the user's .slam_ignore file and
		compile the regular expressions therein. It shouldn't be called directly.
		"""

		self.ignore_dir_patterns_ = []
		ignore_file = os.path.expanduser('~/.slam_ignore')
		if os.path.exists(ignore_file):
			f = open(ignore_file, 'r')
			line_num = 1
			for line in f:
				l = line.strip()
				try:
					self.ignore_dir_patterns_.append(re.compile(l))
				except:
					message.warn('bad regular expression on line ' + str(line_num) + ' of .slam_ignore', [l])
				line_num += 1
	
			f.close()

	def find_configurations_(self, path, env, variants, toolset):
		"""
		Recursively descends in to the path given, looking for and loading any slamfiles that it 
		finds. env, variants and toolset have the same meaning as the constructor arguments of the
		same names.
		
		Used by the constructor, shouldn't be called elsewhere.
		"""
		cfg = os.path.join(path, project.config_name_)
		if os.path.exists(cfg) and os.path.isfile(cfg):
			self.load_profiles_(path, env, variants, toolset, [])

		p = path
		if p == '': p = '.'

		inner_dirs = [d for d in os.listdir(p) if os.path.isdir(os.path.join(path, d))]
		for d in inner_dirs:
			for p in self.ignore_dir_patterns_:
				if p.match(d): break
			else:
				self.find_configurations_(os.path.join(path, d), env, variants, toolset)

	def load_profiles_(self, path, env, variants, toolset, sofar):
		"""
		Loads the slamfile at path/slamfile and follows and 'depends' directives to locate and load
		additional slamfiles.
		
		Used by the constructor, shouldn't be called elsewhere.
		"""
		cfg = os.path.normpath(os.path.join(path, project.config_name_))
		if not self.dependencies_map_.has_key(cfg): self.dependencies_map_[cfg] = []

		if os.path.exists(cfg) and os.path.isfile(cfg):
			for p in self.profiles_:
				if p.get_path() == cfg:
					return p # already parsed and loaded
			else:
				desc = slamfile(cfg)
				prof = profile(desc, env, variants, toolset)
				self.profiles_.append(prof)

			depends = prof.args('depends') or []
			
			for d in depends:
				dprof = self.load_profiles_(os.path.join(path, d), env, variants, toolset, sofar)
				if dprof:
					sofar += self.dependencies_map_[dprof.get_path()] + [dprof]
				#if dprof:
				#	sofar.append(dprof)

			for p in sofar:
				if not p in self.dependencies_map_[cfg]:
					self.dependencies_map_[cfg].append(p)

			return prof
		else:
			if path == '' or path == '.': path = 'this directory'
			raise slam_error('slamfile not found in ' + path)

	def get_profiles(self):
		"""
		Returns a list of profile objects, sorted in to build-order.
		"""
		return self.profiles_


	def get_profile_dependencies(self, profile):
		"""
		Returns a list of profile objects on which the given profile depends.
		"""
		return self.dependencies_map_[profile.get_path()]
