# Copyright 2007 Edd Dawson.
# Distributed under the Boost Software License, Version 1.0.
# See accompanying file BOOST_LICENSE_1_0.txt or copy at http://www.boost.org/LICENSE_1_0.txt

import os
import re
import message
import sys
from sets import Set

# TODO - functionality for splitting up an environment variable in to a list of values?

def warn_deprecated(old, new):
	"""
	Issues a warning message saying that the old switch is deprecated in favour of the new one.
	"""
	message.warn(old + ' is deprecated and will be removed in version 2.3')
	message.warn('please use ' + new + " instead. Type 'slam -i' to view new-style flags")

def get_replacement_arg(old):
	"""
	If old is an argument recognised by slam version 2.1 or below, the new version of the argument
	is returned, else old is returned.
	"""
	replacements = \
	[ \
		('--make-slamfile', '-ms'), \
		('--quiet', '-q'), \
		('--list-ts-flags', '-i'), \
		('--native-deps', '-nd'), \
		('--recursive', '-r'), \
		('--wipe-deps', '-w'), \
		('--dry-run', '-dr'), \
		('--clean', '-c'), \
		('--no-env', '-ne') \
	]
	old = old.strip()
	for r in replacements:
		if old == r[0]:
			warn_deprecated(r[0], r[1])
			return r[1]
	return old

def translate_deprecated(args):
	"""
	Translates any old-style switched recognised by slam version 2.1 or below in the list of 
	strings given by args in to their new equivalents.
	"""
	for i in range(len(args)):
		args[i] = get_replacement_arg(args[i])


class env:
	"""
	Objects of this class represent a slam environment which is obtaned by examining the host
	environment (if requested), the user's .slam_env file (if any) and the command line.
	"""
	def __init__(self, args):
		"""
		Creates an environment from the the user's .slam_env file and args,	which are the 
		command-line arguments given to slam
		"""
		self.env_ = {}
		self.switches_ = []

		translate_deprecated(args)

		if not '-ne' in args:
			self.integrate_host_env()
			
		self.set_variable('cwd', os.getcwd())

		# Apply any defaults found in the user's .slam_defaults file
		self.apply_env_file_(os.path.expanduser('~/.slam_env'), False)

		# Now add any variables or switches from the command line
		for a in args:
			if len(a) == 0: continue
			if '=' in a:
				idx = a.index('=')
				if idx != 0: self.env_[a[:idx].strip()] = a[idx + 1:].strip()
				else: self.switches_.append(a.strip())
			else:
				self.switches_.append(a.strip())
				
		# If an env file was specified, add the settings from there, too
		env_file = self.get_value('env')
		if env_file: self.apply_env_file_(env_file, True)

		norepeats = Set(self.switches_)
		self.switches_ = [s for s in norepeats]
		self.switches_.sort() # So we can e.g. MD5 the environment later on, perhaps
		
	def apply_env_file_(self, filename, warn_when_missing):
		if os.path.exists(filename):
			f = open(filename, 'r')
			for line in f:
				l = line.strip()

				if len(l) == 0: continue

				if '=' in l:
					parts = l.split('=', 1)
					self.env_[parts[0].strip()] = parts[1].strip()
				else:
					self.switches_.append(get_replacement_arg(l.strip()))

			f.close();
		elif warn_when_missing:
			message.warn("couldn't find environment file " + filename)

			
	def get_value(self, variable):
		"""
		Returns the value for the given variable, or None of the variable isn't in the env
		"""
		if variable in self.env_.keys():
			return self.env_[variable]
		return None

	def has_switch(self, sw):
		"""
		Returns True if sw was in the list of switches passed to slam
		"""
		return sw in self.switches_

	def has_variable(self, var):
		"""
		Returns True if var was a variable name in the list of command line arguments passed to slam
		"""
		return var in self.env_.keys()

	def integrate_host_env(self, overwrite = False):
		"""
		Adds the host operating system's environment with this env object
		"""
		for k in os.environ.keys():
			if not k in self.env_.keys() or overwrite:
				self.env_[k] = os.environ[k]

	def set_variable(self, variable, value):
		"""
		Sets the value of a variable in the environment
		"""
		self.env_[variable] = value
		
	def remove_switch(self, sw):
		"""
		Removes a switch from the environment
		"""
		if sw in self.switches_:
			del self.switches_[self.switches_.index(sw)]

	def remove_variable(self, variable):
		"""
		Removes a variable from the environment
		"""
		val = None
		if variable in self.env_.keys():
			val = self.env_[variable]
			del self.env_[variable]
		return val
			

def expand_variables(s, env):
	"""
	Looks for $(XYZ) and @(XYZ) in the string given by s and makes appropriate replacements
	from the given slam.env.env object. The new string with replacements made is returned.
	
	If a "hard" replacement of a particular variable is needed in s that doesn't exist in
	env, a RuntimeError is raised with the name of the variable that couldn't be expanded.
	"""
	def replacement_from_env_(env, match, escape, allow_empty):
		var = match.group()
		prepend = ''

		# remove ?( and ) from the ends
		if var.startswith(escape + escape):
			return var[1:]
		if var.startswith(escape):
			var = var[2:-1]
		else:
			prepend = var[0]
			var = match.group()[3:-1]

		replacement = env.get_value(var)
		if not replacement:
			if not allow_empty:	raise RuntimeError(var)
			replacement = ''
			
		ret = prepend + replacement
		return ret

	rgx = re.compile(r'.?\$\(\w+\)|^\$\(\w+\)')
	phase1 = rgx.sub(lambda m: replacement_from_env_(env, m, '$', False), s)
	
	rgx = re.compile(r'.?@\(\w+\)|^@\(\w+\)')
	phase2 = rgx.sub(lambda m: replacement_from_env_(env, m, '@', True), phase1)

	return phase2
