# Copyright 2007 Edd Dawson.
# Distributed under the Boost Software License, Version 1.0.
# See accompanying file BOOST_LICENSE_1_0.txt or copy at http://www.boost.org/LICENSE_1_0.txt

"""
This module contains all the exception classes specific to slam
"""

import traceback
import sys

class slam_error:
	"""
	This base exception class for slam. All other slam-specific exceptions are derived from this.
	"""
	def __init__(self, what):
		"""
		Creates a slam_error object containing the given string that describes the reason for the
		exception's existence
		"""
		self.what_ = what

	def message(self):
		"""
		Returns a fornatted version of the reason for the exceptions existence
		"""
		return self.what_

	def what(self):
		"""
		Returns the original reason for the exceptions existence
		"""
		return self.what_


class slam_not_implemented_error(slam_error):
	"""
	If an exception of this type is thrown, it means that a particular function being called 
	hasn't been over-ridden in a derived class.
	"""
	def __init__(self, what) :
		slam_error.__init__(self, what)



class slam_cli_error(slam_error):
	"""
	If an exception of this type is thrown, it means that an operation that interfaces with an
	external command-line program has failed.
	"""
	def __init__(self, what) :
		slam_error.__init__(self, what)


class slamfile_error(slam_error):
	"""
	If an exception of this type is thrown, it means that a slamfile could not be parsed.
	Details are contained in the exception object
	"""
	def __init__(self, filename, line, line_num, pos, what):
		"""
		Creates a slamfile_error object that contans details of the parse error in a slamfile.
		The filename, line text, line number, position on the line and a general description
		are to be given to describe the problem.
		"""
		slam_error.__init__(self, what)

		self.filename_ = filename
		self.line_ = line
		self.line_num_ = line_num
		self.pos_ = pos


	def message(self):
		"""
		Returns a formatted error string that describes the nature and position of the error in a
		slamfile
		"""
		msg = 'on line ' + str(self.line_num_) + ' of ' + self.filename_ + ': ' + self.what_ + '\n\n'
		msg += (self.pos_ * ' ') + 'v\n'
		msg += self.line_ # line already has a '\n' on the end
		msg += (self.pos_ * ' ') + '^'
		return msg

	def filename(self):
		"""
		Returns the path of the slamfile in which the error was found
		"""
		return self.filename_

	def line_with_error(self):
		"""
		Returns the text of the line that contains the error
		"""
		return self.line_

	def line_number(self):
		"""
		Returns the number of the line that contains the error
		"""
		return self.line_num_

	def position_on_line(self):
		"""
		Returns the position of the error in the line on which the error occurs
		"""
		return self.pos_


class slam_propagated_exception(slam_error):
	"""
	This exception is used in a threaded scenario. When a thread throws an exception it can be 
	"stored" in a slam_propagated_exception and re-raised when the thread joins its parent thread.
	"""
	def __init__(self):
		exc_info = sys.exc_info()
		self.value_ = exc_info[1]
		self.traceback_ = exc_info[2]

		what = reduce(lambda x,y: x + y, traceback.format_exception(exc_info[0], exc_info[1], exc_info[2]))
		slam_error.__init__(self, what)

	def exception(self):
		"""
		Returns the exception that is being propagated
		"""
		return self.value_

	def traceback(self):
		"""
		Returns the traceback for the exception being propagated
		"""
		return self.traceback_