# Copyright 2007 Edd Dawson.
# Distributed under the Boost Software License, Version 1.0.
# See accompanying file BOOST_LICENSE_1_0.txt or copy at http://www.boost.org/LICENSE_1_0.txt

"""
This module contains the machinery for parsing slamfiles. You should only need to call
parse_slamfile() externally; the other functions are merely functions used in the implementation
of parse_slamfile()
"""

import string
from exceptions import slam_error
from exceptions import slamfile_error

def pluck_token(line, start):
	"""
	Gets a token from the given line (a string) at position start. Returns a tuple containing
	the token aquired and the position of the character at which the token ends.
	"""
	if line[start] == ':': return (':', start + 1)

	n = len(line)
	is_space = lambda c: c in string.whitespace
	element = ''

	while start < n and not is_space(line[start]) and line[start] not in '#:"\'':
		element += line[start]
		start += 1

	return (element, start)


def pluck_quote(line, start, q):
	"""
	Like pluck_token, except that a quotation (delimited at either end by q) is taken from line.
	The quotation is returned in the first element of the tuple without its surrounding quotation
	delimiters.
	"""
	n = len(line)
	element = ''
	escapables = "'\"\\"

	while start < n and line[start] != q:
		if line[start] == '\\':
			start += 1
			if start == n or not line[start] in escapables:
				raise slam_error('Misplaced escape character')

		element += line[start]
		start += 1

	if start == n and n > 0 and line[n-1] != q:
		raise slam_error('Unclosed quotation (' + q + ')')
	else: start += 1

	return (element, start)


def pluck_element(line, start):
	"""
	Delegates to pluck_token or pluck_quote depending on what is found at line[start].
	"""
	n = len(line)

	while start != n and line[start] in string.whitespace:
		start += 1

	if start == n: return ('', start)
	if line[start] in '\'"': return pluck_quote(line, start + 1, line[start])
	elif line[start] == '#': return ('', n)
	else: return pluck_token(line, start)


def parse_slamfile(filename):
	"""
	Parses the slamfile with the given filename. Returns an array of 3-tuples representing the 
	non-empty lines of a slamfile. a slamcore.exceptions.slam_error object is thrown on error
	"""
	f = open(filename, 'rb')

	all_parts = []
	line_number = 1
	i = 0

	for line in f:
		try:
			n = len(line)
			i = 0
			parts = ([], [], [])
			p = 0
			while i != n:
				(element, i) = pluck_element(line, i)
				if len(element) > 0:
					if element == ':':
						p += 1
						if p == 3: raise slam_error('Each line must have 3 parts (directive : variant : arguments)')
					else:
						element = element.strip()
						if element != '':
							parts[p].append(element)

		except slam_error, err:
			f.close()
			raise slamfile_error(filename, line, line_number, i, err.what())

		l = [len(x) for x in parts]
		if sum(l) != 0:
			if l[0] != 1: raise slamfile_error(filename, line, line_number, i, 'directive must have exactly one token')
			if l[1] != 1: raise slamfile_error(filename, line, line_number, i, 'variant must have exactly one token')
			all_parts.append( (parts[0][0], parts[1][0], parts[2]) )

		line_number += 1

	f.close()
	return all_parts