# Copyright 2007 Edd Dawson.
# Distributed under the Boost Software License, Version 1.0.
# See accompanying file BOOST_LICENSE_1_0.txt or copy at http://www.boost.org/LICENSE_1_0.txt

"""
This module contains the task and taskq objects which are used to implement parts of program 
execution that can be run in parallel, such as dependency analysis and some parts of the
actual build process.
"""

import threading
import sys
import message

from slamcore.engine.exceptions import slam_error
from slamcore.engine.exceptions import slam_propagated_exception

class task:
	"""
	A task object is a callable entity with a description.
	"""
	def __init__(self, callback, description = None):
		"""
		Creates a task object from the given callback and description. callback should be a callable
		entity that takes 0 arguments.
		"""
		self.description_ = description
		self.callback_ = callback
		assert callable(self.callback_)

	def __call__(self):
		"""
		Calls the callable entity given at construction time
		"""
		return self.callback_()

	def description(self):
		"""
		Returns the description of the task given at construction time
		"""
		return self.description_

	def __str__(self):
		"""
		Returns a string representation of the task given at construction time, which will be empty
		if None was passed at that time.
		"""
		return self.description_ or ''


class taskpool:
	"""
	A taskpool object represents a pool of "task queues". Each task queue can contain any number of tasks
	that are to be executed sequentially, though all queues in a pool may be processed concurrently.
	"""
	def __init__(self):
		self.mutex_ = threading.Lock()
		self.tasks_ = []
		self.exception_ = None

	def add_queue(self, q):
		"""
		Adds a queue to the pool. q should be a list of task objects
		"""
		self.tasks_.append(q)

	def pluck_queue_(self):
		"""
		Used in the implementation of execute to get a the next un-processed queue from the pool.
		"""
		self.mutex_.acquire()

		n = len(self.tasks_)
		if n > 0:
			seq = self.tasks_[n-1]
			del self.tasks_[n-1]
		else:
			seq = None

		self.mutex_.release()
		return seq

	def per_thread_execute_(self):
		"""
		Used in the implementation of execute. Gets the next un-processed queue and processes it.
		"""
		while True:
			seq = self.pluck_queue_()
			if seq:
				try:
					for cmd in seq:
						if self.dry_run_:
							txt = cmd.description()
							if txt: message.info(txt)
						else:
							cmd()

				except slam_error, ex:
					self.mutex_.acquire()
					if not self.exception_:
						self.exception_ = ex
					self.tasks_ = []
					self.mutex_.release()

				except:
					self.mutex_.acquire()
					if not self.exception_:
						self.exception_ = slam_propagated_exception()
					self.tasks_ = []
					self.mutex_.release()
			else:
				break

	def execute(self, num_threads, dry_run):
		"""
		Processes the queues in the pool using the given number of threads. If dry_run is true, the
		queue is not processed (though it is emptied as is done when dry_run == False), instead the
		descriptions of the tasks are printed to the standard output stream.
		"""
		# Kick off some threads. Each will call per_thread_execute_
		threads = []
		self.dry_run_ = dry_run

		for i in range(num_threads):
			t = threading.Thread(target = self.per_thread_execute_)
			t.start()
			threads.append(t)

		# Wait for all the threads to finish
		for t in threads:
			t.join()

		# If one of the threads encountered an exception, re-throw it here
		if self.exception_:
			self.tasks_ = []
			raise self.exception_