# Copyright 2007 Edd Dawson.
# Distributed under the Boost Software License, Version 1.0.
# See accompanying file BOOST_LICENSE_1_0.txt or copy at http://www.boost.org/LICENSE_1_0.txt

"""
This module contains the definition of the slam_main() function, which organises and 
executes all of the work that slam has to perform
"""

import sys
import os

import slamcore.hlp as hlp
import slamcore.build as build
from slamcore.doc import txt
from slamcore.engine import env
from slamcore.engine import description
from slamcore.engine import deps
from slamcore.engine import message
from slamcore.engine.exceptions import slam_error

def slam_main():
	"""
	This is the entry point for the build-process proper. The order of operations is:
	
	1. setup the slam environment
	2. load the appropriate toolset
	3. analyse the slamfiles found to create a global build description
	4. create and execute build sequence
	"""
	
	# Initialise the environment
	e = env.env(sys.argv[1:])

	# Quit help message and exit if -h or another variant found on the command line
	for h in ['-h', '--help', 'help']:
		if e.has_switch(h):
			hlp.show_help()
			sys.exit(0)

	# Check for quiet mode
	if e.has_switch('-q'): message.set_verbose(False)

	message.info('running version ' + hlp.slam_version())

	# Create a slamfile and exit if -ms on the command line
	if e.has_switch('-ms'):
		hlp.dump_default_slam_file()
		message.info('default slamfile created')
		sys.exit(0)

	# Import the toolset module requested by the user and create a toolset object
	toolset_name = e.get_value('toolset')
	if not toolset_name: raise slam_error('no toolset specified')

	try:
		tspath = e.get_value('tspath')
		if tspath:
			sys.path = [tspath] + sys.path
			exec 'from ' + toolset_name + ' import create as create_toolset'
		else:
			exec 'from slamcore.toolsets.' + toolset_name + ' import create as create_toolset'
	except ImportError:
		raise slam_error("no toolset named '" + toolset_name + "' found")

	ts = create_toolset(e)

	# If the user wanted info on the toolset (-i), print it and exit
	if e.has_switch('-i'):
		print ts.notes(txt.formatter())
		sys.exit(0)

	# Find out the variant requested by the user
	variant = e.get_value('variant')
	if not variant: raise slam_error('no variant specified')
	
	# Create the build description (project object) by analysing the slamfiles specified
	recurse = e.has_switch('-r')
	message.info('looking for slamfiles')
	if variant == 'common':
		proj = description.project(e, ['common'], ts, recurse)
	else:
		proj = description.project(e, [variant, 'common'], ts, recurse)

	# Load the dependencies database, or create a new one if it doesn't exist
	db = deps.database('slam.db')
	if e.has_switch('-w'):
		message.info('wiping dependencies for variant=' + variant + ', toolset=' + ts.name())
		db.wipe_deps(proj, ts)
		
	# Find out how many threads the user wants to use, default to 1
	num_threads = e.get_value('threads')
	if num_threads:
		try:
			num_threads = int(num_threads)
		except:
			raise slam_error('number of threads must be an integer!')

		if num_threads < 1:
			raise slam_error('number of threads must be positive!')
	else:
		num_threads = 1;

	# If the user specifies a 'dry run' with -dr, we don't actually build anything,
	# instead we simply echo what would happen.
	dry_run = e.has_switch('-dr')
	
	# Update the dependencies database if we're not 'cleaning' or 'tidying'
	if not (e.has_switch('-c') or e.has_switch('-t')):
		message.info('checking source level dependencies')
		db.update(proj, ts, num_threads)
		if not dry_run: db.save()
	
	# Generate the the ordered list of taskpools of commands to execute in order to
	# carry out the build
	if e.has_switch('-c'): gen_queue = build.generate_clean_queue
	elif e.has_switch('-t'): gen_queue = build.generate_tidy_queue
	else: gen_queue = build.generate_build_queue
	
	message.info('generating build sequence')
	queue = gen_queue(db, proj, ts)
	
	# Begin the build!
	message.info('running build sequence')
	for pool in queue:
		pool.execute(num_threads, dry_run)

	if dry_run: message.info('done (dry run)')
	else: message.info('done')