# Copyright 2007 Edd Dawson.
# Distributed under the Boost Software License, Version 1.0.
# See accompanying file BOOST_LICENSE_1_0.txt or copy at http://www.boost.org/LICENSE_1_0.txt

import sys
from sets import Set

import slamcore.toolsets.gcc as gcc
import slamcore.toolsets.toolset as toolset
import slamcore.toolsets.utils as utils
from slamcore.engine.taskpool import task
from slamcore.engine import message

# http://developer.apple.com/releasenotes/DeveloperTools/RN-GCC4/index.html


always = toolset.always

class applegcc(gcc.gcc):
	def __init__(self, env):
		gcc.gcc.__init__(self, env, 'lib', '.a', 'lib', '.dylib', '', '', '', '.o')

		self.ub_ = env.has_switch('ub')
		self.native_deps_ = env.has_switch('-nd')
		self.arch_flags_ = ('-arch', 'i386', '-arch', 'ppc')

		if sys.byteorder == 'big': message.info('applegcc toolset thinks this is a Power-PC machine')
		else: message.info('applegcc toolset thinks this is an Intel machine')

		if self.ub_: message.info('applegcc toolset is building universal binaries')

		if self.ub_:
			self.map_flag(always(), compiler = self.arch_flags_, linker = self.arch_flags_)

		self.set_identifier('applegcc', 4)

		desc = \
		"""
		This toolset uses Apple's version of g++ developed for Darwin and Mac OS X
		to build C++ libraries and programs.

		It is capable of building universal binaries. See the 'Special features'
		section.

		This toolset is provided with the slam distribution.
		"""

		features = \
		"""
		If you put 'ub' on slam's command line when using this toolset, object code
		will be generated in universal binary format, otherwise the target processor
		will be that of the machine doing the compilation.

		This toolset also recognises the 'frameworks' and 'framework-dirs' directives
		so that applications can link against libraries within Mac OS X frameworks
		and use their headers.

		For example, to link against the OpenGL framework, you would add the
		following line to your 	slamfile:

		frameworks : common : OpenGL # line in slamfile

		The frameworks directive is 'inherited' through the 'depends' directive, in
		the same way as libs.

		The frameworkdirs directive allows you to specify custom search locations
		for Mac OS X frameworks. The frameworkdirs directive is NOT inherited via
		the depends directive.
		"""

		issues = \
		"""
		It is desirable to have Mac OS X 'Bundle' capabilities. Hopefully such a
		feature can be implemented in such a way that there is some degree of
		uniformity between Mac OS X bundles and Windows resources, for example.
		"""

		more =\
		"""
		Information specific to Apple's GCC:

		http://developer.apple.com/tools/xcode/
		http://developer.apple.com/documentation/DeveloperTools/gcc-4.0.1/gcc/Invoking-GCC.html
		"""

		self.replace_note('Toolset description', desc)
		self.add_note('Special features', features)
		self.add_note('Known issues', issues)
		self.add_note('More information', more)


	def librarian(self, static):
		if static: return ['libtool', '-static']
		else:
			raise slam_not_implemented_error("dynamic libraries are not yet supported by slam's applegcc toolset")
			return ['libtool', '-dynamic'] # TODO

	def librarian_output_args(self, profile):
		return ['-o', self.target_product(profile)]

	def lib_tasks(self, profile, profile_deps, static):
		flags = self.librarian_flags(profile)
		flags.update(toolset.tool_flags(self, profile, 'librarian'))

		cmd = 	toolset.args_to_list(self.librarian(static)) + \
				toolset.args_to_list(self.librarian_output_args(profile)) + \
				toolset.args_to_list(flags) + \
				toolset.args_to_list(self.librarian_input_args(profile))

		ret = []
		d = profile.args('dir')
		if d: ret.append(task(lambda d_=d[0]: utils.ensure_dir_exists(d_)))
		ret.append(toolset.cmdline_task(cmd))

		return ret

	def frameworks(self, profiles):
		fwks = Set()
		for prof in profiles: fwks.update(prof.args('frameworks') or [])

		ret = []
		for f in fwks: ret += ['-framework', f]

		return ret

	def framework_dirs(self, profile):
		fwkdirs = Set()
		fwkdirs.update(profile.args('frameworkdirs') or [])

		ret = []
		for f in fwkdirs: ret += ['-F', f]

		return ret

	def linker_input_args(self, profile, profile_deps):
		fwks = self.frameworks([profile] + profile_deps)
		return gcc.gcc.linker_input_args(self, profile, profile_deps) + fwks

	def input_args(self, filename, profile):
		fwkdirs = self.framework_dirs(profile)
		return gcc.gcc.input_args(self, filename, profile) + fwkdirs

	def native_deps_cmd(self, filename, profile):
		defs = self.defines_args(profile)
		incs = self.include_args(profile)
		flags = self.compiler_flags(profile)
		flags.update(toolset.tool_flags(self, profile, 'compiler'))

		if self.ub_ and self.native_deps_:
			# Need to remove -arch flags from dependency generation command,
			# for some reason Apple's g++ doesn't like it.
			flags = flags.difference([self.arch_flags_])

		flags = toolset.args_to_list(flags)
		return toolset.cmdline(self.compiler() + flags + ['-MM', filename] + defs + incs)

def create(env):
	return applegcc(env)
