# Copyright 2007 Edd Dawson.
# Distributed under the Boost Software License, Version 1.0.
# See accompanying file BOOST_LICENSE_1_0.txt or copy at http://www.boost.org/LICENSE_1_0.txt

import os

import slamcore.toolsets.cpp as cpp
import slamcore.toolsets.toolset as toolset
import slamcore.toolsets.utils as utils
from slamcore.engine.exceptions import slam_not_implemented_error
from slamcore.engine.taskpool import task
from slamcore.engine import message

has_some = toolset.has_some
has_all = toolset.has_all
has_none = toolset.has_none
has_not = toolset.has_not
always = toolset.always

# VERY useful: http://www.objectcentral.com/vide/help/videdoc/bcc32.html
# Note that we can get very close to a native dependencies calculation
# facility by using -m and related flags. However, I cannot find out how
# to supress object file generation :(

class borland(cpp.cpp):
	def __init__(self, env):
		cpp.cpp.__init__(self, env, '', '.lib', '', '.dll', '', '.exe', '', '.obj')

		self.map_flag(flag = always(), compiler = ['-q', '-c', '-w-inl', '-w-ext', '-w-8066', '-w-amp', '-w-8080', '-RT', '-x', '-xd', '-Jgd'], linker = '-q', librarian = '/P32')

		self.map_flag(flag = 'strict', compiler = '-w')
		self.map_flag(flag = 'warnings-as-errors', compiler = '-w!')

		self.map_flag(flag = 'as-c++', compiler = '-P')
		self.map_flag(flag = 'debug', compiler = ['-R', '-v', '-y', '-xp', '-D_HAS_ITERATOR_DEBUGGING=1'], linker = '-lv')

		self.map_flag(flag = 'optimize-off', compiler = '-Od')
		self.map_flag(flag = 'optimize-min', compiler = '-O')
		self.map_flag(flag = 'optimize', compiler = '-O2')
		self.map_flag(flag = 'optimize-max', compiler = ['-O2', '-Ox', '-Z'])
		self.map_flag(flag = 'optimize-size', compiler = ['-O1', '-d'])
		self.map_flag(flag = 'optimize-frame-pointer', compiler = '-k-')
		self.map_flag(has_not('optimize-floating-point'), compiler = '-ff-')


		self.map_flag(flag = 'gui', linker = '-W')
		self.map_flag(flag = 'threads', linker = '-WM')
		self.map_flag(flag = has_all('console', has_not('gui')), linker = '-WC')

		self.set_identifier('borland', 3)

		desc = \
		"""
		This toolset uses Borland's bcc32 compiler to build C++ libraries and
		programs. Versions 5.5.1 and 5.82 have been tested.

		Version 5.5.1 came with Borland's "free command line tools" package,
		which was superceded by version 5.82. 5.82 comes as part of the new
		Turbo C++ package, released in 2006.

		This toolset is provided with the slam distribution.
		"""

		deps = \
		"""
		This toolset does not have native dependency analysis facilities;
		slam's default analysis tool is used.
		"""

		issues = \
		"""
		From the VIDE manual (see More information for URL):
		The standard Windows header files included with Borland BCC 5.5
		(e.g. windows.h) are set up to compile Windows applications for Windows
		2000/XP. This is NOT documented anywhere in the help files that come with
		BCC 5.5. There are some significant differences in some Win32 API calls
		for Win2K/XP and earlier Win95, Win98, and WinNT 4.0 versions.
		Code compiled for Win2K/XP (which is the default) will not run correctly
		on older versions of Windows. Code compiled for older versions will run
		on Win2K/XP.

		You should consider adding the following 2 arguments to the defines
		directive in your slamfile if you want your GUI code to work on Windows
		versions prior to Win2K!

		defines : common : WINVER=0x0400 _WIN32_WINNT=0x0400

		Mapping for the 'ansi' and 'iso' slam flags were removed starting from
		version 3 as they caused more trouble than they were worth.
		"""

		# ^ Is the above WINVER stuff true for all Windows compilers???

		more = \
		"""
		Borland C++ Compiler version 5.5 Free download:
		http://www.codegear.com/tabid/139/Default.aspx

		Turbo C++ 2006 free download:
		http://www.turboexplorer.com/cpp

		This page of the VIDE manual was incredibly helpful:
		http://www.objectcentral.com/vide/help/videdoc/bcc32.html
		"""

		self.add_note('Toolset description', desc)
		self.add_note('Dependency analysis', deps)
		self.add_note('Known issues', issues)
		self.add_note('More information', more)

	def build_tasks(self, filename, profile):
		flags = self.compiler_flags(profile)
		flags.update(toolset.tool_flags(self, profile, 'compiler'))

		cmd = 	toolset.args_to_list(self.compiler()) + \
				toolset.args_to_list(flags) + \
				toolset.args_to_list(self.defines_args(profile) + self.include_args(profile)) + \
				toolset.args_to_list(self.output_args(filename, profile)) + \
				toolset.args_to_list(self.input_args(filename, profile))

		ret = []
		d = profile.args('dir')
		if d: ret.append(task(lambda d_=d[0]: utils.ensure_dir_exists(d_)))
		ret.append(toolset.cmdline_task(cmd))

		return ret

	def link_tasks(self, profile, profile_deps):
		flags = self.linker_flags(profile)
		flags.update(toolset.tool_flags(self, profile, 'linker'))

		libdirs = self.libs([profile] + profile_deps, getlibs = False)

		cmd = 	toolset.args_to_list(self.linker()) + \
				toolset.args_to_list(flags) + \
				toolset.args_to_list(libdirs) + \
				toolset.args_to_list(self.linker_output_args(profile)) + \
				toolset.args_to_list(self.linker_input_args(profile, profile_deps))

		ret = []
		d = profile.args('dir')
		if d: ret.append(task(lambda d_=d[0]: utils.ensure_dir_exists(d_)))
		ret.append(toolset.cmdline_task(cmd))

		return ret

	def lib_tasks(self, profile, profile_deps, static):
		product = self.target_product(profile)
		tasks = [task(lambda p=product: utils.remove_file(p), 'removing existing library file')]
		tasks += cpp.cpp.lib_tasks(self, profile, profile_deps, static)
		return tasks

	def clean_product_tasks(self, profile):
		tasks = []
		build_type = toolset.argument(profile, 'type')

		if build_type and build_type == 'executable':
			tdsfile = os.path.splitext(self.target_product(profile))[0] + '.tds'
			tasks += [task(lambda f=tdsfile: utils.remove_file(f), 'removing ' + tdsfile)]

		return tasks + cpp.cpp.clean_product_tasks(self, profile)

	def compiler(self):
		return ['bcc32']

	def linker(self):
		return ['bcc32']

	def librarian(self, static):
		if static: return ['tlib']
		else: raise slam_not_implemented_error("dynamic libraries are not yet supported by slam's borland toolset")

	def input_args(self, filename, profile):
		return [filename]

	def output_args(self, filename, profile):
		return ['-o' + self.build_product(filename, profile)]

	def librarian_input_args(self, profile):
		return ['+' + obj for obj in self.object_files(profile)]

	def librarian_output_args(self, profile):
		return [self.target_product(profile)]

	def linker_input_args(self, profile, profile_deps):
		profs = [profile] + profile_deps
		objects = self.object_files(profile)
		libs = self.libs(profs, getdirs = False)
		return objects + libs

	def linker_output_args(self, profile):
		return ['-e' + self.target_product(profile)]

	def defines_args(self, profile):
		defines = profile.args('defines') or []
		defs = []
		for d in defines: defs += ['-D' + d]
		return defs

	def include_args(self, profile):
		includes = profile.args('include') or []
		incs = []
		for i in includes: incs += ['-I' + i]
		return incs

	def format_libdir(self, libdir):
		return ['-L' + libdir]

	def format_lib(self, lib):
		return [lib + '.lib']


def create(env):
	return borland(env)
