# Copyright 2007 Edd Dawson.
# Distributed under the Boost Software License, Version 1.0.
# See accompanying file BOOST_LICENSE_1_0.txt or copy at http://www.boost.org/LICENSE_1_0.txt

"""
This module contains the cpp class, which implements the slamcore.toolsets.toolset interface to
provide a base class for all C++ toolsets.
"""

import os
import copy

import slamcore.toolsets.toolset as toolset
import slamcore.toolsets.includes as includes
import slamcore.toolsets.utils as utils
from slamcore.engine.message import warn
from slamcore.engine.exceptions import slam_not_implemented_error
from slamcore.engine.exceptions import slam_error
from slamcore.engine.taskpool import task


class cpp(toolset.toolset):
	"""
	This is the base class for all C++ toolsets. It provides narrower abstractions for the compiler,
	librarian and linker components of a C++ tool chain.
	"""
	def __init__(self, env, lib_prefix, lib_suffix, dylib_prefix, dylib_suffix, exe_prefix, exe_suffix, obj_prefix, obj_suffix):
		"""
		Creates a new cpp base object that builds static libraries, dynamic libraries, executables and object files with the
		supplied prefixes and suffixes.
		"""
		toolset.toolset.__init__(self)
		self.env_ = copy.deepcopy(env)

		self.fix_ = \
		{ \
			'executable' : [exe_prefix, exe_suffix], \
			'static-library' : [lib_prefix, lib_suffix], \
			'dynamic-library' : [dylib_prefix, dylib_suffix], \
			'object' : [obj_prefix, obj_suffix] \
		}

		notice = \
		"""
		This toolset is experimental. If you find anything wrong with it, please
		report the problem at http://www.mr-edd.co.uk
		"""

		self.add_note('Notice', notice)

	# ---------------------------------------------------------------------------
	# functions usable in derived classes (and possibly in other client code)
	# ---------------------------------------------------------------------------

	def env(self):
		"""
		Returns the environment used to construct the toolset.
		"""
		return self.env_

	def compiler_flags(self, profile):
		"""
		Returns a sets.Set object containing the flags to be used with the compiler.
		"""
		flags = self.get_flag_args(profile, 'compiler')
		flags.update(toolset.tool_flags(self, profile, 'compiler'))
		return flags

	def linker_flags(self, profile):
		"""
		Returns a sets.Set object containing the flags to be used with the linker.
		"""
		flags = self.get_flag_args(profile, 'linker')
		flags.update(toolset.tool_flags(self, profile, 'linker'))
		return flags

	def librarian_flags(self, profile):
		"""
		Returns a sets.Set object containing the flags to be used with the librarian.
		"""
		flags = self.get_flag_args(profile, 'librarian')
		flags.update(toolset.tool_flags(self, profile, 'librarian'))
		return flags

	def map_flag(self, flag, compiler = None, linker = None, librarian = None):
		"""
		maps the string of decision predicate given by flag on to the command line argument(s) for
		a compiler and/or linker and/or librarian.

		compiler, linker and librarian may each be a string or a tuple, or a (heterogeneous) list
		thereof.
		"""
		self.add_flag('compiler', flag, compiler)
		self.add_flag('linker', flag, linker)
		self.add_flag('librarian', flag, librarian)

	def unmap_flag(self, flag, compiler = None, linker = None, librarian = None):
		"""
		Removes the mapping for the compiler and/or linker and/or librarian. The arguments have the
		same meaning as those in map_flag()
		"""
		self.rem_flag('compiler', flag, compiler)
		self.rem_flag('linker', flag, linker)
		self.rem_flag('librarian', flag, librarian)

	def default_deps(self, filename, profile):
		"""
		Returns a slamcore.engine.taskpool.task object, complete with description, that when called
		returns a list of dependencies for the source file identified by filename that is controlled
		by the slamcore.engine.description.profile object given by profile.
		"""
		incs = profile.args('include') or []
		defs = profile.args('defines') or []
		gather = lambda f=filename, i=incs, d=defs: includes.get_includes(f, i, d)
		hsh = ''
		for i in incs: hsh += ' i#' + i
		for d in defs: hsh += ' d#' + d
		return task(gather, '<slam deps analyser> [' + hsh + ' ]')

	# ---------------------------------------------------------------------------
	# overrides from the toolset base class
	# ---------------------------------------------------------------------------

	def deps(self, filename, profile):
		"""
		Overrides the function from the toolset base class. If the toolset has native dependency
		calculation abilities and the '-nd' flag was passed to slam, the native dependency system
		will be used by the returned task, else default_deps will be returned.
		"""
		native = self.env().has_switch('-nd')
		if native: return self.native_deps(filename, profile)
		else: return self.default_deps(filename, profile)

	def build_tasks(self, filename, profile):
		"""
		Overrides the function from the toolset baseclass.

		Classes derived from this one need to implement compiler(), output_args(), input_args(),
		defines_args() and include_args() in order to customise the behaviour of this function.
		"""
		flags = self.compiler_flags(profile)

		cmd = 	toolset.args_to_list(self.compiler()) + \
				toolset.args_to_list(self.output_args(filename, profile)) + \
				toolset.args_to_list(flags) + \
				toolset.args_to_list(self.input_args(filename, profile)) + \
				toolset.args_to_list(self.defines_args(profile) + self.include_args(profile))

		ret = []
		d = profile.args('dir')
		if d: ret.append(task(lambda d_=d[0]: utils.ensure_dir_exists(d_)))
		ret.append(toolset.cmdline_task(cmd))

		return ret

	def build_product(self, filename, profile):
		"""
		Overrides the function from the toolset baseclass.
		"""
		dirs = profile.args('dir') or ['']
		if len(dirs) > 1: warn('output directory for ' + profile.get_path() + ' has multiple values, using ' + dirs[0])

		filename = os.path.split(filename)[1]
		filename = os.path.splitext(filename)[0]
		filename = self.fix_['object'][0] + filename + self.fix_['object'][1]

		return os.path.join(toolset.argument(profile, 'dir') or '', filename)

	def target_tasks(self, profile, profile_deps):
		"""
		Overrides the function from the toolset baseclass.

		Classes derived from this one need to implement lib_tasks() and link_tasks() in order to
		customise the behaviour of this function.
		"""
		typ = toolset.argument(profile, 'type')
		if typ == 'dynamic-library': return self.lib_tasks(profile, profile_deps, False)
		elif typ == 'static-library': return self.lib_tasks(profile, profile_deps, True)
		elif typ == 'executable': return self.link_tasks(profile, profile_deps)
		else: warn("unrecognised target type '" + typ + "' in " + profile.get_path() + ', no target will be built')

	def target_product(self, profile):
		"""
		Overrides the function from the toolset baseclass.
		"""
		build_type = toolset.argument(profile, 'type')
		target = toolset.argument(profile, 'target')

		if build_type and target:
			if '/' in target or '\\' in target:
				raise slam_error('/ or \\ found in argument to target directive in ' + profile.get_path())

			fixes = self.fix_.get(build_type) or ['', '']
			return os.path.join(toolset.argument(profile, 'dir') or '', fixes[0] + target + fixes[1])

		return None

	def clean_tasks(self, filename, profile):
		"""
		Overrides the function from the toolset baseclass.
		"""
		product = self.build_product(filename, profile)
		if product:
			desc = 'removing ' + product
			return [task(lambda p=product: utils.remove_file(p), desc)]
		else:
			return []

	def clean_product_tasks(self, profile):
		"""
		Overrides the function from the toolset baseclass.
		"""
		product = self.target_product(profile)
		if product:
			desc = 'removing ' + product
			return [task(lambda p=product: utils.remove_file(p), desc)]
		else:
			return []

	# ---------------------------------------------------------------------------
	# implementation functions, may be overriden in derived toolsets
	# ---------------------------------------------------------------------------

	def native_deps(self, filename, profile):
		"""
		By default returns self.default_deps(filename, profile) and issues a warning about the
		toolset not having a native dependency analysis facility.

		Classes derived from this one should reimplement this function if they do provide such
		a facility.
		"""
		warn('toolset ' + self.name() + ' does not have native dependency calculation facilities; using default method')
		return self.default_deps(filename, profile)

	def compiler(self):
		"""
		Derived classes need to implement this function in order to return a list of strings
		containing the name of the compiler e.g. ['g++'] for GCC's g++. The list is necessary in
		order to provide support for tools such as ccache or distcc in future.
		"""
		raise slam_not_implemented_error("compiler is not implemented for this toolset")

	def linker(self):
		"""
		Derived classes need to implement this function in order to return a list of strings
		containing the name of the linker.
		"""
		raise slam_not_implemented_error("linker is not implemented for this toolset")

	def librarian(self, static):
		"""
		Derived classes need to implement this function in order to return a list of strings
		containing the name of the librarian.
		"""
		raise slam_not_implemented_error("librarian is not implemented for this toolset")

	def input_args(self, filename, profile):
		"""
		Derived classes need to implement this function in order to return a list of strings
		that form the arguments needed to specify to the compiler that filename is the source file
		to be compiled according to profile (a slamcore.engine.description.profile object)
		"""
		raise slam_not_implemented_error("input_args is not implemented for this toolset")

	def output_args(self, filename, profile):
		"""
		Derived classes need to implement this function in order to return a list of strings that
		form the arguments needed to specify the name of the object file created by the	compilation
		of filename according to profile (a slamcore.engine.description.profile object).
		"""
		raise slam_not_implemented_error("output_args is not implemented for this toolset")

	def librarian_input_args(self, profile):
		"""
		Derived classes need to implement this function in order to return a list of strings that
		form the arguments needed to specify the input objects to the librarian according to
		profile (a slamcore.engine.description.profile object)
		"""
		raise slam_not_implemented_error("librarian_input_args is not implemented for this toolset")

	def librarian_output_args(self, profile):
		"""
		Derived classes need to implement this function in order to return a list of strings that
		form the arguments needed to specify the output library to the librarian according to
		profile (a slamcore.engine.description.profile object)
		"""
		raise slam_not_implemented_error("librarian_output_args is not implemented for this toolset")

	def linker_input_args(self, profile, profile_deps):
		"""
		Derived classes need to implement this function in order to return a list of strings that
		form the arguments needed to specify the input objects and libraries to the linker
		according to profile (a slamcore.engine.description.profile object)
		"""
		raise slam_not_implemented_error("linker_input_args is not implemented for this toolset")

	def linker_output_args(self, profile):
		"""
		Derived classes need to implement this function in order to return a list of strings that
		form the arguments needed to specify the output executable to the linker according to
		profile (a slamcore.engine.description.profile object)
		"""
		raise slam_not_implemented_error("linker_output_args is not implemented for this toolset")

	def defines_args(self, profile):
		"""
		Derived classes need to implement this function in order to return a list of strings that
		form the arguments needed to specify the pre-defined pre-processor macros according to
		profile (a slamcore.engine.description.profile object)
		"""
		raise slam_not_implemented_error("defines_args is not implemented for this toolset")

	def include_args(self, profile):
		"""
		Derived classes need to implement this function in order to return a list of strings that
		form the arguments needed to specify the #include search-paths according to profile
		(a slamcore.engine.description.profile object)
		"""
		raise slam_not_implemented_error("include_args is not implemented for this toolset")

	def format_libdir(self, libdir):
		"""
		Derived classes need to implement this function in order to return a list of strings that
		form the arguments needed to specify that libdir is a library search-path.
		"""
		raise slam_not_implemented_error("format_libdir is not implemented for this toolset")

	def format_lib(self, lib):
		"""
		Derived classes need to implement this function in order to return a list of strings that
		form the arguments needed to specify that lib is a library to be linked.
		"""
		raise slam_not_implemented_error("format_lib is not implemented for this toolset")

	def lib_tasks(self, profile, profile_deps, static):
		"""
		Classes derived from this one need to implement librarian(), librarian_output_args() and
		librarian_input_args() in order to customise the behaviour of this function.
		"""
		flags = self.librarian_flags(profile)

		cmd = 	toolset.args_to_list(self.librarian(static)) + \
				toolset.args_to_list(flags) + \
				toolset.args_to_list(self.librarian_output_args(profile)) + \
				toolset.args_to_list(self.librarian_input_args(profile))

		ret = []
		d = profile.args('dir')
		if d: ret.append(task(lambda d_=d[0]: utils.ensure_dir_exists(d_)))
		ret.append(toolset.cmdline_task(cmd))

		return ret

	def link_tasks(self, profile, profile_deps):
		"""
		Classes derived from this one need to implement linker(), linker_output_args() and
		linker_input_args() in order to customise the behaviour of this function.
		"""
		flags = self.linker_flags(profile)

		cmd = 	toolset.args_to_list(self.linker()) + \
				toolset.args_to_list(flags) + \
				toolset.args_to_list(self.linker_output_args(profile)) + \
				toolset.args_to_list(self.linker_input_args(profile, profile_deps))

		ret = []
		d = profile.args('dir')
		if d: ret.append(task(lambda d_=d[0]: utils.ensure_dir_exists(d_)))
		ret.append(toolset.cmdline_task(cmd))

		return ret

	# ---------------------------------------------------------------------------
	# convenience functions
	# ---------------------------------------------------------------------------

	def object_files(self, profiles):
		"""
		Returns the list of object files that result from a succesful build of profile, which is a
		slamcore.engine.description.profile object.
		"""
		if not type(profiles) is list: profiles = [profiles]
		ret = []
		for p in profiles:
			sources = p.args('sources') or []
			ret += [self.build_product(s, p) for s in sources]

		return ret

	def libs(self, profiles, getlibs = True, getdirs = True, rem_repeats = True):
		"""
		Returns a list of arguments that specify the libraries and their search paths if getlibs and
		getdirs are True, respectively.

		If rem_repeats is true, repeat specifications of the libraries and library search paths are
		removed.
		"""
		ret = []
		profs = profiles[:]
		profs.reverse()

		for p in profs:
			if getdirs:
				libdirs = p.args('libdirs') or []
				libdirs = [os.path.normpath(os.path.join(p.get_dir(), d)) for d in libdirs]
				libdirs = [self.format_libdir(d) for d in libdirs]

				# If there are elements in libdirs that are already in ret, remove them from libdirs before appending
				if rem_repeats: libdirs = [d for d in libdirs if d not in ret]
				ret += libdirs

			if getlibs:
				libraries = p.args('libs') or []
				libraries = [self.format_lib(l) for l in libraries]

				# If there are elements in ret that are also in libraries remove them from ret before appending
				if rem_repeats: ret = [l for l in ret if l not in libraries]
				ret += libraries

			ptype = toolset.argument(p, 'type')
			if ptype == 'static-library' or ptype == 'dynamic-library':
				if getdirs:
					d = p.args('dir')
					if d:
						d = self.format_libdir(d[0])
						if (not rem_repeats) or (not d in ret): ret.append([d])

				if getlibs:
					lib = toolset.argument(p, 'target')
					if lib != '':
						lib = self.format_lib(lib)
						if rem_repeats: ret = [l for l in ret if l != lib]
						ret.append([lib])

		# Currently, ret is a list of lists. We need to flatten it in to a '1D' list:
		flat_ret = []
		for l in ret:
			flat_ret += l

		return flat_ret

