# Copyright 2007 Edd Dawson.
# Distributed under the Boost Software License, Version 1.0.
# See accompanying file BOOST_LICENSE_1_0.txt or copy at http://www.boost.org/LICENSE_1_0.txt

import os
import threading

import slamcore.toolsets.cpp as cpp
import slamcore.toolsets.toolset as toolset
import slamcore.toolsets.utils as utils
from slamcore.engine.exceptions import slam_not_implemented_error
from slamcore.engine.taskpool import task

always = toolset.always
has_some = toolset.has_some

def flatten(lst):
	if isinstance(lst, basestring): return [lst]
	ret = []
	for e in lst: ret += flatten(e)
	return ret

class dm(cpp.cpp):
	def __init__(self, env, lib_prefix, lib_suffix, dylib_prefix, dylib_suffix, exe_prefix, exe_suffix, obj_prefix, obj_suffix):
		cpp.cpp.__init__(self, env, lib_prefix, lib_suffix, dylib_prefix, dylib_suffix, exe_prefix, exe_suffix, obj_prefix, obj_suffix)

		self.map_flag(always(), compiler = ['-Ae', '-Ab', '-Ar', '-mn'])

		self.map_flag(has_some('iso', 'ansi'), compiler = ['-A', '-D_STLP_NO_NATIVE_WIDE_FUNCTIONS=1'])
		self.map_flag(flag = 'strict', compiler = ['-w-', '-Aw'])
		self.map_flag(flag = 'warnings-as-errors', compiler = '-wx')
		self.map_flag(flag = 'as-c++', compiler = '-cpp')
		self.map_flag(flag = 'debug', compiler = ['-g', '-gt', '-gh', '-gp', '-S'], librarian = ['-p128'])

		# self.map_flag(flag = 'optimize-off', compiler = '-Jb')
		# ^^^ causes executables to crash, probably due to layout
		# differences in standard library binary

		self.map_flag(flag = 'optimize', compiler = '-o')
		self.map_flag(flag = 'optimize-size', compiler = '-o+space')

		self.set_identifier('dm', 4)

		desc = \
		"""
		This toolset uses Walter Bright's Digital Mars C++ compiler to build
		libraries and programs for the Windows platform. It is recommended
		that you use the dmc-specific build of STLPort with this toolset,
		available from the Digital Mars website.

		If you also have a version of Microsoft's Visual C++ compiler
		installed on your system, make sure you do NOT invoke builds using
		the dm toolset when the visual studio environment is enabled (e.g.
		when using the "Visual Studio Command Prompt")

		This is because the Digital Mars and Microsoft linkers have the same
		name (link.exe) and the changes made to the environment will cause
		the Microsoft linker to be selected over the Digital Mars linker,
		causing all kinds of strange errors to occur should the build attempt
		to link an executable.

		This toolset is provided with the slam distribution.
		"""

		deps = \
		"""
		This toolset does not have native dependency analysis facilities;
		slam's default analysis tool is used.
		"""

		issues = \
		"""
		Unless you have the full Digital Mars STLPort (from the Digital Mars
		CD), you shouldn't add DEBUG to the defines directive, as this
		tells the linker to use a specific version of the STLPort library
		binary. However, DEBUG is present as an argument to the defines in
		the default slamfile, so you should remove it unless you either:

		* copy lib/stlp45dm_static.lib to lib/stlp45dm_stldebug_static.lib
		* hack stlport/stlport/config/stl_dm.h

		This isn't slam's fault; defining DEBUG as a pre-processor symbol is
		quite sensible as far as I'm concerned and it's a shame the default
		Digital Mars installation makes this awkward.

		I am still investigating a possible implementation of a dependency
		analysis system using facilities native to dmc.

		Some versions of the Digital Mars linker are prone to deadlock on
		multicore systems. If you experience problems, set the process
		affinity of Digital Mars' link.exe to a particular core.

		The following features still need to be implemented:
		* dynamic library support
		* precompiled header support
		"""

		more = \
		"""
		For more information on Digital Mars, visit:
		http://www.digitalmars.com/

		In particular, the following page is useful when investigating dmc's
		various options: http://www.digitalmars.com/ctg/sc.html
		"""

		self.add_note('Toolset description', desc)
		self.add_note('Dependency analysis', deps)
		self.add_note('Known issues', issues)
		self.add_note('More information', more)

	#def native_deps(self, filename, profile):
		# N.B:
		# dmc has a -d compiler flag that alleges to create a .dep file.
		# This file doesn't seem to create a lot of useful information that could be used
		# to implement native_deps. I also can't figure out how to stop dmc generating the
		# object file at the same time.
		# Need to investigate this further

	def compiler(self):
		return ['dmc']

	def linker(self):
		return ['link']

	def librarian(self, static):
		if static: return ['lib', '-c', '-n']
		else:
			raise slam_not_implemented_error("dynamic libraries are not yet supported by slam's dm toolset")
			return ['dmc'] # dmc is used to assemble dlls, when we get around to implementing it

	def input_args(self, filename, profile):
		return ['-c', filename]

	def output_args(self, filename, profile):
		return ['-o' + self.build_product(filename, profile)]

	def librarian_input_args(self, profile):
		return self.object_files(profile)

	def librarian_output_args(self, profile):
		return [self.target_product(profile)]

	def link_tasks(self, profile, profile_deps):
		flags = self.linker_flags(profile)

		cmd = toolset.args_to_list(self.linker())

		obj = '+'.join(self.object_files(profile))
		out = self.target_product(profile)
		map = ''
		lib = '+'.join(flatten(self.libs([profile] + profile_deps)))
		cmd.append(','.join([obj, out, map, lib]))

		ret = []
		d = profile.args('dir')
		if d: ret.append(task(lambda d_=d[0]: utils.ensure_dir_exists(d_)))
		ret.append(toolset.cmdline_task(cmd))

		return ret


	def defines_args(self, profile):
		defines = profile.args('defines') or []
		defs = []
		for d in defines: defs += ['-D' + d]
		return defs

	def include_args(self, profile):
		includes = profile.args('include') or []
		incs = []
		for i in includes: incs += ['-I' + i]
		return incs

	def format_libdir(self, libdir):
		libdir = os.path.normpath(libdir)
		if not libdir.endswith('\\'): libdir += '\\'
		return libdir

	def format_lib(self, lib):
		return [lib + '.lib']

def create(env):
	return dm(env, '', '.lib', '', '.dll', '', '.exe', '', '.obj')
