# Copyright 2007 Edd Dawson.
# Distributed under the Boost Software License, Version 1.0.
# See accompanying file BOOST_LICENSE_1_0.txt or copy at http://www.boost.org/LICENSE_1_0.txt

import os
import re

import slamcore.toolsets.cpp as cpp
import slamcore.toolsets.toolset as toolset
import slamcore.toolsets.utils as utils
from slamcore.engine.exceptions import slam_cli_error
from slamcore.engine.exceptions import slam_not_implemented_error
from slamcore.engine.taskpool import task
from slamcore.engine import message

has_some = toolset.has_some
non_escaped_space = re.compile(r'([^\\]) ')
version_rgx = re.compile(r'(\d)\.(\d)\.(\d)(.*)', re.M)

def gcc_version(compiler):
	try:
		m = version_rgx.match(utils.get_cmd_output(compiler + ' -dumpversion'))
		if m: return (int(m.group(1)), int(m.group(2)), int(m.group(3)), m.group(4))
	except:
		pass

	return (0, 0, 0)

def split_depends_line(line):
	deps = non_escaped_space.sub('\\g<1>\n', line).split('\n')
	return [d.replace(r'\ ', ' ') for d in deps]



class gcc(cpp.cpp):
	def __init__(self, env, lib_prefix, lib_suffix, dylib_prefix, dylib_suffix, exe_prefix, exe_suffix, obj_prefix, obj_suffix):
		cpp.cpp.__init__(self, env, lib_prefix, lib_suffix, dylib_prefix, dylib_suffix, exe_prefix, exe_suffix, obj_prefix, obj_suffix)

		self.gcc_exe = (env.get_value('g++') or 'g++').split()

		c = self.compiler()[0]
		v = gcc_version(c)

		if v[0] == 0:
			message.info('gcc toolset could not determine the version of GCC installed')
		else:
			vs = reduce(lambda x,y: str(x) + '.' + str(y), v[:3])
			message.info('gcc toolset determined that you are using a variant of g++ version ' + vs)

		self.map_flag(flag = has_some('iso', 'ansi'), compiler = ['-ansi', '-pedantic'])
		self.map_flag(flag = 'warnings-as-errors', compiler = '-Werror')
		self.map_flag(flag = 'debug', compiler = ['-ggdb3', ('-D', '_GLIBCXX_DEBUG'), ('-D', '_GLIBCXX_DEBUG_PEDANTIC')])
		self.map_flag(flag = 'strict', compiler = ['-Wall', '-Wswitch'])
		self.map_flag(flag = 'as-c++', compiler = '-xc++')

		self.map_flag(flag = 'optimize-off', compiler = '-O0')
		self.map_flag(flag = 'optimize-min', compiler = '-O1')
		self.map_flag(flag = 'optimize', compiler = '-O2')
		self.map_flag(flag = 'optimize-max', compiler = '-O3')
		self.map_flag(flag = 'optimize-size', compiler = '-Os')
		self.map_flag(flag = 'optimize-loops', compiler = '-funroll-loops')
		self.map_flag(flag = 'optimize-frame-pointer', compiler = '-fomit-frame-pointer')
		self.map_flag(flag = 'optimize-floating-point', compiler = '-ffast-math')

		if v[0] > 3:
			message.info('enabling flags specific to gcc 4')
			self.map_flag(flag = 'optimize-vector', compiler = '-ftree-vectorize')

		if v[0] > 3 or (v[0] == 3 and v[1] > 3):
			# -W renamed to -Wextra in gcc 3.4
			self.map_flag(flag = 'strict', compiler = '-Wextra')
		else:
			self.map_flag(flag = 'strict', compiler = '-W')

		self.set_identifier('gcc', 3)

		desc = \
		"""
		This toolset uses GCC's g++ compiler to build C++ libraries and programs.
		You should be able to use it on any system that has a g++ installation,
		though more specialised toolsets exist for some platforms based on this
		one.

		This toolset attempts to detect the version of gcc that you have installed.
		Some flag mappings are dependent on the version of gcc found.

		This toolset is provided with the slam distribution.
		"""

		deps = \
		"""
		This toolset uses g++ itself with the -MM switch to analyse dependencies
		for each source file if the -nd flag is passed to slam.
		"""

		issues = \
		"""
		There are no known issues with the facilities provided by this toolset.

		The following features still need to be implemented, however:

		* dynamic library support
		* precompiled header support
		* support for distcc
		* support for ccache, though it shouldn't be needed if slam works properly!
		"""

		features = \
		"""
		If 'strip' is present in the arguments to the flags directive of a slamfile
		that builds an executable, the build product will be stripped of all its
		symbols. This is only recommended for release builds since the binary will
		no longer contain the information necessary for debugging. For example,
		adding a line such as the following to a slamfile would be the recommended
		course of action:

		flags : release : strip

		This toolset supports pkg-config through the slam's pkgconfig directive.
		Simply add the names of the packages as arguments to this directive.

		You can specify the name of the g++ exectable by adding g++=<name> on the
		slam command line. With this feature you can use ccache with slam by adding
		"g++=ccache g++" to the slam command line.
		"""

		more = \
		"""
		For more information on GCC and g++, visit: http://gcc.gnu.org/
		Direct link to manuals: http://gcc.gnu.org/onlinedocs/
		"""

		self.add_note('Toolset description', desc)
		self.add_note('Dependency analysis', deps)
		self.add_note('Special features', features)
		self.add_note('Known issues', issues)
		self.add_note('More information', more)

	def native_deps_cmd(self, filename, profile):
		defs = self.defines_args(profile)
		incs = self.include_args(profile)
		flags = self.compiler_flags(profile)
		flags.update(toolset.tool_flags(self, profile, 'compiler'))
		flags = toolset.args_to_list(flags)
		return toolset.cmdline(self.compiler() + flags + ['-MM', filename] + defs + incs)

	def gather_native_deps(self, filename, profile):
		cmd = self.native_deps_cmd(filename, profile)
		try:
			deps = utils.get_cmd_output(cmd)
			if ':' in deps:
				idx = deps.index(':')

				# Sometimes gcc can put multiple dependencies on the same line
				# We have to recognise "\ " as a space in a filename, as opposed to " ",
				# which is a space between two adjacent files
				lines = [d.strip() for d in deps[idx + 1:].split('\\\n')]
				deps = []
				for line in lines:
					deps += [os.path.normpath(d) for d in split_depends_line(line)]
				return deps
		except slam_cli_error, err:
			# Got an error from g++.
			# It will fail again on compilation, so the user will have to fix their code
			# at some point. Thus returning [] here is delaying the inevitable, at worst :)
			return []

	def native_deps(self, filename, profile):
		gather = lambda f=filename, p=profile: self.gather_native_deps(f, p)
		return task(gather, self.native_deps_cmd(filename, profile))

	def compiler(self):
		return self.gcc_exe

	def linker(self):
		return self.gcc_exe

	def librarian(self, static):
		if static: return ['ar']
		else: raise slam_not_implemented_error("dynamic libraries are not yet supported by slam's gcc toolset")

	def input_args(self, filename, profile):
		return ['-c', filename]

	def output_args(self, filename, profile):
		return ['-o', self.build_product(filename, profile)]

	def librarian_input_args(self, profile):
		return self.object_files(profile)

	def librarian_output_args(self, profile):
		return ['-rs', self.target_product(profile)]

	def linker_input_args(self, profile, profile_deps):
		profs = [profile] + profile_deps
		objects = self.object_files(profile)
		libs = self.libs(profs)
		return objects + libs

	def linker_output_args(self, profile):
		return ['-o', self.target_product(profile)]

	def strip_exe_tasks(self, profile):
		return [toolset.cmdline_task(['strip', '-s', self.target_product(profile)])]

	def link_tasks(self, profile, profile_deps):
		tasks = cpp.cpp.link_tasks(self, profile, profile_deps)
		flags = profile.args('flags') or []
		if 'strip' in flags: tasks += self.strip_exe_tasks(profile)
		return tasks

	def defines_args(self, profile):
		defines = profile.args('defines') or []
		defs = []
		for d in defines: defs += ['-D', d]
		return defs

	def include_args(self, profile):
		includes = profile.args('include') or []
		incs = []
		for i in includes: incs += ['-I', i]
		return incs

	def format_libdir(self, libdir):
		return ['-L' + libdir]

	def format_lib(self, lib):
		return ['-l' + lib]


def create(env):
	exe_ext = ''
	if os.name == 'nt': exe_ext = '.exe'
	return gcc(env, 'lib', '.a', 'lib', '.so', '', exe_ext, '', '.o')
