# Copyright 2007 Edd Dawson.
# Distributed under the Boost Software License, Version 1.0.
# See accompanying file BOOST_LICENSE_1_0.txt or copy at http://www.boost.org/LICENSE_1_0.txt

"""
This module contains facilities for scanning C++ source files for #include directives.
"""

import re
import os
from slamcore.engine import message

block_comment_ = r'\s*(/\*.*\*/\s*)*'
includes_ = \
[	\
	block_comment_ + r'#' + block_comment_ + 'include' + block_comment_ + r'"\s*([\w/\\\-\. ]*)\s*"',	\
	block_comment_ + r'#' + block_comment_ + 'include' + block_comment_ + r'<\s*([\w/\\\-\. ]*)\s*>'	\
]

locals_rgx_ = re.compile(includes_[0])
system_rgx_ = re.compile(includes_[1])


def get_header_from_line(line, include_dirs, locals_only):
	"""
	Given a line of text from a C++ source file, and a list of directories in which to look for
	headers, this function will return the path of a header if a #include directive was found on
	this line, else it will return None.
	
	If locals_only is True, then only #include directives that surround the header name in double
	quotes will be analysed (names surrounded by angle brackets are ignored).
	"""
	line = line.lstrip()

	match = locals_rgx_.match(line)
	if match:
		header = match.group(4)
		for d in include_dirs:
			filename = os.path.realpath(os.path.join(d, header))
			if os.path.exists(filename) and os.path.isfile(filename):
				return filename

		message.warn("couldn't resolve header " + header + " in any of the following directories", include_dirs)
		return None

	if not locals_only:
		match = system_rgx_.match(line)
		if match:
			header = match.group(4)
			for d in include_dirs:
				filename = os.path.realpath(os.path.join(d, header))
				if os.path.exists(filename) and os.path.isfile(filename):
					return filename

	return None


def get_includes_for_file(filename, headers_so_far, include_dirs, defines, locals_only):
	"""
	Returns a list of paths to header files that are #included by filename, or any one of its 
	#included headers, or their #included headers and so on.
	
	include_dirs should specify the paths to be used in order to attempt to resolve header names.
	
	Currently defines are ignored, but may be used if we can implement something closer to a proper
	C++ pre-processor in future.
	
	headers_so_far contains a list of the headers that we have so far determined have been #included
	(the implementation is recursive).

	If locals_only is True, then only #include directives that surround the header name in double
	quotes will be analysed (names surrounded by angle brackets are ignored).
	"""
	f = open(filename, 'rb')

	d = os.path.realpath(os.path.split(filename)[0])
	incs = include_dirs[:]
	if not d in incs: incs = [d] + incs

	for line in f:
		hdr = get_header_from_line(line, incs, locals_only)

		if hdr and hdr not in headers_so_far:
			headers_so_far.append(hdr)
			headers_so_far = get_includes_for_file(hdr, headers_so_far, include_dirs, defines, locals_only)

	f.close()

	return headers_so_far


def get_includes(filename, include_dirs, defines, locals_only = False):
	"""
	Returns a list of headers that are #included by filename and all of its headers (recursively).
	
	include_dirs is a list containing the directories in which to attempt to resolve header names.
	
	defines is a list that contains all the preprocessor macros that are defined during the build.
	
	If locals_only is True, then only #include directives that surround the header name in double
	quotes will be analysed (names surrounded by angle brackets are ignored).
	"""
	full_dirs = [os.path.realpath(d) for d in include_dirs]
	deps = get_includes_for_file(filename, [], full_dirs, defines, locals_only)
	deps.sort()
	return deps