# Copyright 2007 Edd Dawson.
# Distributed under the Boost Software License, Version 1.0.
# See accompanying file BOOST_LICENSE_1_0.txt or copy at http://www.boost.org/LICENSE_1_0.txt

import os

import slamcore.toolsets.cpp as cpp
import slamcore.toolsets.toolset as toolset
import slamcore.toolsets.utils as utils
from slamcore.engine.exceptions import slam_not_implemented_error
from slamcore.engine.taskpool import task

has_some = toolset.has_some
has_all = toolset.has_all
has_none = toolset.has_none
has_not = toolset.has_not
always = toolset.always

class msvc(cpp.cpp):
	def __init__(self, env):
		cpp.cpp.__init__(self, env, '', '.lib', '', '.dll', '', '.exe', '', '.obj')

		self.map_flag(flag = always(), compiler = '/nologo', linker = '/nologo', librarian = '/nologo')

		self.map_flag(flag = has_some('iso', 'ansi'), compiler = ['/EHsc', '/GR'])
		self.map_flag(flag = 'strict', compiler = ['/W3'])
		self.map_flag(flag = 'warnings-as-errors', compiler = '/WX')
		self.map_flag(flag = 'as-c++', compiler = '/TP')
		self.map_flag(flag = 'debug', compiler = '/Zi')

		self.map_flag(flag = 'optimize-off', compiler = '/Od')
		self.map_flag(flag = 'optimize-min', compiler = '/Ot')
		self.map_flag(flag = 'optimize', compiler = '/O2')
		self.map_flag(flag = 'optimize-max', compiler = ['/Ox', '/Oi'])
		self.map_flag(flag = 'optimize-size', compiler = ['/Os', '/Gf', '/GF'])
		self.map_flag(flag = 'optimize-frame-pointer', compiler = '/Oy')
		self.map_flag(flag = 'optimize-floating-point', compiler = '/Op-')

		self.map_flag(flag = has_all('dynamic-runtime', 'debug'), compiler = '/MDd')
		self.map_flag(flag = has_all('dynamic-runtime', has_not('debug')), compiler = '/MD')
		self.map_flag(flag = has_all(has_not('dynamic-runtime'), 'debug'), compiler = '/MTd')
		self.map_flag(flag = has_none('dynamic-runtime', 'debug'), compiler = '/MT')
		self.map_flag(flag = 'gui', compiler = '/GA')

		self.set_identifier('msvc', 2)

		desc = \
		"""
		This toolset uses Microsoft's Visual C++ compiler to build C++ libraries
		and programs. You should be able to use it on any system that has an
		installation of Microsoft Visual C++.

		This toolset will pick up any version of Visual C++ that you have
		installed. To use features specific to a particular version of msvc,
		please see the toolsets derived from this one e.g. msvc6, msvcdotnet and
		msvc8. It is generally better to use one of those rather than this one.

		When using this toolset, you should make sure that the host environment
		is set up correctly, usually done by running the "Visual Studio Command
		Prompt", found somewhere in the Start menu.

		Standard C++ exception handling and run-time type identification (RTTI)
		are turned on by default with this toolset and those derived from it.
		To disable them, you should derive your own toolset and customise to suit.
		This is much easier than it sounds!

		This toolset is provided with the slam distribution.
		"""

		deps = \
		"""
		This toolset does not have native dependency analysis facilities;
		slam's default analysis tool is used.
		"""

		issues = \
		"""
		If you use the pkgconfig directive in your slamfiles, you may
		occasionally see warnings such as:

		"slam warning: removing unsupported compiler flag '-mms-bitfields'
		wrongly suggested by pkg-config"

		This is due to a fault with the compiler flags suggested by pkg-config
		whereby gcc specific flags are suggested for use with Visual C++. It is
		probably safe to ignore, though I am not 100% sure. It is a problem
		with pkg-config, not with slam.

		The following features still need to be implemented:

		* dynamic library support
		* precompiled header support
		* global optimisation support
		* support for linked-in resources
		"""

		features = \
		"""
		This toolset supports pkg-config through the slam's pkgconfig directive.
		Simply add the names of the packages as arguments to this directive.

		Microsoft compilers allow you to link your applications against either a
		system-wide runtime library, or a library specific to your application,
		so that may be more easily upgraded or is resilient to change, depending
		on how you look at it! The default behaviour is to link against the
		system-wide runtime library.

		To link against an application specific runtime, add 'dynamic-runtime'
		as an argument to the flags directive in your slamfiles.
		"""

		more = \
		"""
		For more information on Microsoft Visual Studio and Visual C++, visit:
		http://www.msdn.com/
		http://msdn2.microsoft.com/en-us/library/f35ctcxw.aspx
		"""

		self.add_note('Toolset description', desc)
		self.add_note('Dependency analysis', deps)
		self.add_note('Special features', features)
		self.add_note('Known issues', issues)
		self.add_note('More information', more)

	def compiler(self):
		return ['cl']

	def linker(self):
		return ['link']

	def librarian(self, static):
		if static: return ['lib']
		else: raise slam_not_implemented_error("dynamic libraries are not yet supported by slam's msvc toolset")

	def input_args(self, filename, profile):
		return ['/c', filename]

	def output_args(self, filename, profile):
		return ['/Fo' + self.build_product(filename, profile)]

	def librarian_input_args(self, profile):
		return self.object_files(profile)

	def librarian_output_args(self, profile):
		return ['/OUT:' + self.target_product(profile)]

	def linker_input_args(self, profile, profile_deps):
		profs = [profile] + profile_deps
		objects = self.object_files(profile)
		libs = self.libs(profs)
		return objects + libs

	def linker_output_args(self, profile):
		return ['/OUT:' + self.target_product(profile)]

	def defines_args(self, profile):
		defines = profile.args('defines') or []
		defs = []
		for d in defines: defs += ['/D', d]
		return defs

	def include_args(self, profile):
		includes = profile.args('include') or []
		incs = []
		for i in includes: incs += ['/I', i]
		return incs

	def format_libdir(self, libdir):
		return ['/LIBPATH:' + libdir]

	def format_lib(self, lib):
		return [lib + '.lib']


def create(env):
	return msvc(env)
