# Copyright 2007 Edd Dawson.
# Distributed under the Boost Software License, Version 1.0.
# See accompanying file BOOST_LICENSE_1_0.txt or copy at http://www.boost.org/LICENSE_1_0.txt

import os
from sets import Set

import slamcore.toolsets.msvc as msvc
import slamcore.toolsets.toolset as toolset
import slamcore.toolsets.utils as utils
from slamcore.engine.taskpool import task
from slamcore.engine.exceptions import slam_cli_error

# base-class for msvc-based toolsets other than msvc6
class msvcdotnet(msvc.msvc):
	def __init__(self, env):
		msvc.msvc.__init__(self, env)
		
		self.map_flag(flag = 'debug', compiler = ['/RTCu', '/RTCc', '/RTCs'], linker = '/DEBUG')
		self.map_flag(toolset.always(), compiler = ('/D', '_CRT_SECURE_NO_DEPRECATE'))

		self.set_identifier('msvcdotnet', 2)
		
		desc = \
		"""
		This toolset uses the 2003 edition of Microsoft's Visual C++ compiler
		to build C++ libraries and programs. It is also used as a base for
		msvc8 toolset (which uses the 2005 Visual C++ compiler) so please see
		the msvc8 toolset if you wish to use features of Visual C++ 2005.

		When using this toolset, you should make sure that the host environment
		is set up correctly, usually done by running the "Visual Studio Command
		Prompt", found somewhere in the Start menu.

		Standard C++ exception handling and run-time type identification (RTTI)
		are turned on by default with this toolset and those derived from it.

		Similarly, the _SCL_SECURE_NO_DEPRECATE preprocessor macro is defined on
		all builds because I believe it's silly not to have it (you shouldn't
		let kids play with knives). Of course this is a matter of taste;

		To change the default behaviours, you should derive your own toolset and
		customise to suit. This is much easier than it sounds!

		This toolset is provided with the slam distribution.
		"""

		deps = \
		"""
		This toolset uses cl itself with the /showIncludes and /Zs switches to
		analyse dependencies for each source file if the -nd flag is passed to 
		slam.
		"""

		self.replace_note('Toolset description', desc)
		self.replace_note('Dependency analysis', deps)

	def native_deps_cmd(self, filename, profile):
		defs = self.defines_args(profile)
		incs = self.include_args(profile)
		flags = self.compiler_flags(profile)
		flags.update(toolset.tool_flags(self, profile, 'compiler'))
		flags = toolset.args_to_list(flags)

		return toolset.cmdline(['cl', '/showIncludes', '/Zs', filename] + defs + incs + flags)

	def gather_native_deps(self, filename, profile):
		cmd = self.native_deps_cmd(filename, profile)
		try:
			lines = utils.get_cmd_output(cmd).split('\n');
			lines = [ln[22:].strip() for ln in lines if ln.startswith('Note: including file:')]
			deps = Set([os.path.normpath(ln) for ln in lines])
			return [d for d in deps]

		except slam_cli_error, err:
			# Got an error from cl.
			# It will fail again on compilation, so the user will have to fix their code
			# at some point. Thus returning [] here is delaying the inevitable, at worst :)
			return []

	def native_deps(self, filename, profile):
		gather = lambda f=filename, p=profile: self.gather_native_deps(f, p)
		return task(gather, self.native_deps_cmd(filename, profile))
		
	def clean_product_tasks(self, profile):
		product = self.target_product(profile)
		if product:
			tasks = []
			(base, ext) = os.path.splitext(product)
			for e in ['.ilk', '.pdb', ext]:
				desc = 'removing ' + base + e
				tasks += [task(lambda p=base+e: utils.remove_file(p), desc)]
			return tasks
		else:
			return []
		
def create(env):
	return msvcdotnet(env)