# Copyright 2007 Edd Dawson.
# Distributed under the Boost Software License, Version 1.0.
# See accompanying file BOOST_LICENSE_1_0.txt or copy at http://www.boost.org/LICENSE_1_0.txt

"""
This module contains the toolset base class for all other toolsets, which should reimplement
its methods to suit.

This module also contains the associated machinery for setting up flags-to-command-line arguments
mappings for the programs that a toolset uses (e.g. compiler, linker, librarian for most C++
toolsets).

Flags mapping is done using "decision predicates". A decision predicate is a callable object that
takes a single container (set/list/dict etc) argument and returns a boolean value.

Numerous decision predicate types are exposed by this module. Many of them are designed to be
composed with other decision predicates. All decision predicates should have a __str__ method.
"""

from sets import Set

from slamcore.engine.taskpool import task
from slamcore.engine.message import warn
from slamcore.engine.exceptions import slam_not_implemented_error
import slamcore.toolsets.utils as utils


class has:
	"""
	has is a decision predicate; x = has('xyz') creates an object x for which
	x(container) returns True if 'xyz' is in container.
	"""
	def __init__(self, flag): self.flag_ = flag
	def __call__(self, flags): return self.flag_ in flags
	def __str__(self): return self.flag_


class has_not:
	"""
	has_not is a decision predicate; x = has_not('xyz') creates an object x for which
	x(container) returns True if 'xyz' is not in container.
	"""
	def __init__(self, flag):
		if type(flag) is str: self.flag_ = has(flag)
		else: self.flag_ = flag

	def __call__(self, flags):
		return not self.flag_(flags)

	def __str__(self):
		return "not(" + str(self.flag_) + ')'


class has_all(object):
	"""
	has_all is a decision predicate; x = has_all(p1, ..., pN) creates an object x for which
	x(container) returns True if all of p1 to pN are in container.

	Think "and".
	"""
	def __init__(self, *flags):
		"""
		Each argument given may either be a string or another decision predicate. Each string given
		is internally converted in to a has decision predicate.
		"""
		self.elements_ = []
		for f in flags:
			if type(f) is str: self.elements_.append(has(f))
			else: self.elements_.append(f)

	def __call__(self, flags):
		for e in self.elements_:
			if not e(flags): return False
		return True

	def __str__(self):
		join = lambda x,y: str(x) + ' and ' + str(y)
		return '(' + reduce(join, self.elements_) + ')'


class has_some(object):
	"""
	has_some is a decision predicate; x = has_some(p1, ..., pN) creates an object x for which
	x(container) returns True if at least one of p1 to pN is in container.

	Think "or".
	"""
	def __init__(self, *flags):
		"""
		Each argument given may either be a string or another decision predicate. Each string given
		is internally converted in to a has decision predicate.
		"""
		self.elements_ = []
		for f in flags:
			if type(f) is str: self.elements_.append(has(f))
			else: self.elements_.append(f)

	def __call__(self, flags):
		for e in self.elements_:
			if e(flags): return True
		return False

	def __str__(self):
		join = lambda x,y: str(x) + ' or ' + str(y)
		return '(' + reduce(join, self.elements_) + ')'


class has_none(object):
	"""
	has_none is a decision predicate; x = has_none(p1, ..., pN) creates an object x for which
	x(container) returns True if none of p1 to pN are in container.
	"""
	def __init__(self, *flags):
		"""
		Each argument given may either be a string or another decision predicate. Each string given
		is internally converted in to a has decision predicate.
		"""
		self.elements_ = []
		for f in flags:
			if type(f) is str: self.elements_.append(has(f))
			else: self.elements_.append(f)

	def __call__(self, flags):
		for e in self.elements_:
			if e(flags): return False
		return True

	def __str__(self):
		join = lambda x,y: str(x) + ' or ' + str(y)
		return 'not(' + reduce(join, self.elements_) + ')'

class always(object):
	"""
	always is a decision predicate; x = always() creates an object x for which always(container)
	returns True, regardless of the contents of container.
	"""
	def __init__(self): pass
	def __call__(self, flags): return True
	def __str__(self): return '<default>'


def args_to_list(flags_sequence):
	"""
	Expands a sequence of flags in to a list. The flags_sequence given can contain strings and
	tuples. The elements of each tuple found are kept next to each other in the	expansion and
	are expanded such that the resulting list is "1D".

	For example, args_to_list(sets.Set(['abc', ('1', '2', '3')])) may return either
	['abc', '1', '2', '3'] or ['1', '2', '3', 'abc'], depending on how the set is represented.
	The elements in the tuple remain contiguous in the expansion in either case, however.
	"""
	ret = []
	for f in flags_sequence:
		if type(f) is tuple: ret += [t for t in f]
		else: ret.append(f)

	return ret

class toolset(object):
	"""
	This is the base class for all of slam's toolsets. It defines the interface that all derived
	classes should strive to implement.

	See also slamcore.toolsets.cpp.cpp, which is the base class for all C++ toolsets.

	Each toolset module should also provide a create() function (at module scope) that takes a
	slamcore.engine.env.env object as its only argument to return the derived toolset.
	"""

	def __init__(self):
		self.flags_ = {}
		self.names_ = []
		self.name_ = None
		self.notes_ = []
		self.version_ = 0
		self.versions_ = []


	# ---------------------------------------------------------------------------
	# toolset identity
	# ---------------------------------------------------------------------------
	def set_identifier(self, name, version):
		"""
		Sets the name of the toolset (string) and its version (int). The constructors of derived
		toolsets should call this.
		"""
		self.name_ = name
		self.version_ = version
		self.names_ += [name]
		self.versions_ += [version]

	def name(self):
		"""
		Returns the name of the toolset, which should be set by calling set_identifier() in each
		derived toolset's constructor.
		"""
		return self.name_

	def names(self):
		"""
		Returns a list of names containing the name of this toolset and of each toolset that is a
		base class of this one.
		"""
		return self.names_

	def version(self):
		"""
		Returns the version of the toolset, which should be set by calling set_identifier() in each
		derived toolset's constructor.
		"""
		return self.version_

	def identifiers(self):
		"""
		Returns a list of (name, version) tuples of this toolset and of each toolset that is a
		base class of this one.
		"""
		return zip(self.names_, self.versions_)


	# ---------------------------------------------------------------------------
	# flag mapping facilities
	# ---------------------------------------------------------------------------
	def add_flag(self, tool, flag_test, values):
		"""
		Adds a flag to command-line switch mapping for the toolset. tool is a string containing the
		name of the tool for which the mapping applies. For example, this could be 'compiler' or
		'linker' in a C++ toolset.

		flag_test is a decision predicate and values contains the arguments that should be used with
		the tool should the decision predicate yield true for a particular list of flags.

		values can be a list, a string or a tuple of strings. If it is a string, a mapping from the
		decision predicate to the argument given by value is made.

		If values is a tuple then a mapping from the decision predicate to the elements of the tuple
		is made and the elements of a tuple will remain adjacent when expanded for use in a command
		line.

		If values is a string, the net effect is equivalent to calling
		add_flag(self, tool, flag_test, v) for each v in values.
		"""
		if not values: return

		if type(flag_test) is str: flag_test = has(flag_test)
		flag = str(flag_test)

		if not self.flags_.has_key(tool): self.flags_[tool] = {}
		if not self.flags_[tool].has_key(flag): self.flags_[tool][flag] = (flag_test, Set())

		if not type(values) is list: self.flags_[tool][flag][1].add(values)
		else: self.flags_[tool][flag][1].update(values)

	def rem_flag(self, tool, flag_test, values):
		"""
		Removes the mapping that may have been created by a call to
		add_flag(self, tool, flag_test), values)
		"""
		if not values: return

		if type(flag_test) is str: flag_test = has(flag_test)
		flag = str(flag_test)

		if not self.flags_.has_key(tool) or not self.flags_[tool].has_key(flag): return

		if not type(values) is list: self.flags_[tool][flag][1].discard(values)
		else: self.flags_[tool][flag][1].difference_update(values)

	def get_flag_args(self, profile, tool):
		"""
		Returns a sets.Set object of strings and string-tuples that contains the command-line
		switches to be used when running the named tool in the building of the sources associated
		with the given slamcore.engine.description.profile object.
		"""
		if not self.flags_.has_key(tool): return Set()

		flags = [f for f in Set(profile.args('flags') or [])]
		ret = Set()

		for f in self.flags_[tool]:
			(use, values) = self.flags_[tool][f]
			if use(flags):
				ret.update(values)

		return ret


	# ---------------------------------------------------------------------------
	# documentation facilities
	# ---------------------------------------------------------------------------
	def add_note(self, heading, body):
		"""
		Adds some documentation to the toolset with the given heading and body.
		"""
		for i in range(len(self.notes_)):
			if self.notes_[i][0] == heading:
				self.notes_[i][1] += '\n\n' + body
				break
		else:
			self.notes_.append([heading, body])

	def replace_note(self, heading, body):
		"""
		If a note with the given heading already exists in the toolset's documentation, the body is
		replaced with that given. Otherwise, it is as if add_note(self, heading, body) is called.
		"""
		for i in range(len(self.notes_)):
			if self.notes_[i][0] == heading:
				self.notes_[i] = [heading, body]
				break
		else:
			self.add_note(heading, body)

	def remove_note(self, heading):
		"""
		If a note with the given heading is found, it is removed.
		"""
		for i in range(len(self.notes_)):
			if self.notes_[i][0] == heading:
				del self.notes_[i]
				break

	def flags_notes_(self, formatter):
		"""
		Used in the implementation of notes(). Returns a string fomatted using formatter that
		contains information about the flags mappings for this toolset
		"""
		def reduce_if_tuple(t, sep = ' '):
			if type(t) is tuple: return reduce(lambda x,y,s=sep: x + s + y, t)
			else: return t

		ret = ''
		for tool in self.flags_:
			txt = []
			for flag in self.flags_[tool]:
				values = self.flags_[tool][flag][1]
				if len(values):
					lst = ''
					for v in values: lst += ' ' + reduce_if_tuple(v)
					txt.append('- ' + flag + ' ->' + lst + '\n')

			if len(txt) > 0:
				txt.sort()
				txt = ''.join(txt)
				ret += formatter.heading(tool + ' flags mapping') + formatter.text(txt)

		return ret

	def base_notes_(self, formatter):
		"""
		Used in the implementation of notes(). Returns a string fomatted using formatter that
		contains information about this toolset's the base toolsets and their versions.
		"""
		ret = ''
		identifiers = self.identifiers()
		if len(identifiers) > 1:
			identifiers = identifiers[:-1]
			identifiers.reverse()
			ret += formatter.heading('Base toolsets')
			ret += formatter.text("'" + self.name() + "' version " + str(self.version()) + ' is derived from the following toolsets:')
			txt = ''

			for (n, v) in identifiers:
				txt += '* ' + n + ' v' + str(v) + '\n'

			ret += formatter.text(txt)

		return ret

	def notes(self, formatter):
		"""
		Returns a string, created with the help of a formatter (see slamcore.engine.doc.txt for
		example) containing information on this toolset. This is shown when you call slam with the
		'-i' switch on the command line.
		"""
		ret = formatter.preamble()
		ret += formatter.title("Information on toolset '" + self.name() + "'")
		ret += formatter.heading('Version') + formatter.text(str(self.version()))

		for n in self.notes_:
			ret += formatter.heading(n[0]) + formatter.text(n[1])

		ret += self.base_notes_(formatter)
		ret += self.flags_notes_(formatter)
		ret += formatter.postamble()

		return ret


	# ---------------------------------------------------------------------------
	# main facilities
	# ---------------------------------------------------------------------------

	def deps(self, filename, profile):
		"""
		Derived classes should implement this function in order to return a
		slamcore.engine.taskpool.task object that when called returns the dependencies for
		the source file with the given filename that is built according to the instructions in the
		slamcore.engine.description.profile object given by profile.

		The description of the returned task must be an identifier for the process of gathering
		dependencies for the given file. Typically it will be a command line.

		Each time dependencies are checked, slam looks tosee if this description has changed from
		the last time the project was built. If the description has changed, then the dependencies
		for filename will be regenerated.
		"""
		raise slam_not_implemented_error("deps is not implemented for this toolset")

	def build_tasks(self, filename, profile):
		"""
		Derived classes should implement this function in order to return a list of
		slamcore.engine.taskpool.task objects that when executed in order, build the target
		associated with the source file identified by filename, whose build is controlled by the
		slamcore.engine.description.profile object given by profile.
		"""
		raise slam_not_implemented_error("build_command is not implemented for this toolset")

	def build_product(self, filename, profile):
		"""
		Derived classes should implement this function in order to return the name of the file that
		is built by the tasks returned from	build_tasks(self, filename, profile).
		"""
		raise slam_not_implemented_error("build_products is not implemented for this toolset")

	def clean_tasks(self, filename, profile):
		"""
		Derived classes should implement this function in order to remove any files made during the
		execution of the tasks returned by build_tasks(self, filename, profile).
		"""
		raise slam_not_implemented_error("clean_tasks is not implemented for this toolset")

	def target_tasks(self, profile, profile_deps):
		"""
		Derived classes should implement this function in order to return a list of
		slamcore.engine.taskpool.task objects that when executed in order build the final product
		for profile and the profiles on which it depends, given by profile_deps.

		profile and the elements of the profile_deps list are slamcore.engine.description.profile
		objects.
		"""
		raise slam_not_implemented_error("target_tasks is not implemented for this toolset")

	def target_product(self, profile):
		"""
		Derived classes should implement this function in order to return the name of the final
		product that is built by the tasks returned from target_tasks(self, profile).

		profile is an slamcore.engine.description.profile object.
		"""
		raise slam_not_implemented_error("target_product is not implemented for this toolset")

	def clean_product_tasks(self, profile):
		"""
		Derived classes should implement this function in order to remove any files made during the
		execution of the tasks returned by a corresponding call to target_tasks()
		"""
		raise slam_not_implemented_error("clean_product_tasks is not implemented for this toolset")


def cmdline(args_list):
	"""
	Takes a sequence of strings and string-tuples of command line arguments and reduces it to a
	string, which is returned.
	"""
	cmd = ''
	for a in args_list:
		if type(a) is str: sub_args = [a]
		else: sub_args = a

		for s in sub_args:
			if ' ' in s: cmd += '"' + s + '" '
			else:
				s = s.strip()
				if len(s) > 0: cmd += s + ' '

	if len(args_list): return cmd[:-1] # remove trailing space

	return cmd

def cmdline_task(args_list):
	"""
	Creates a slamcore.taskpool.task object that executes the process given by the command-line
	returned from cmdline(args_list). The task returned will have an appropriate description.
	"""
	cmd = cmdline(args_list)
	return task(lambda c=cmd: utils.execute_cmd(c), 'calling ' + cmd)


def argument(profile, directive):
	"""
	Given a slamcore.engine.descrition.profile object and the name of a directive, the first
	argument is returned. If the directive has more than one argument, a warning is issued.

	If the directive didn't exist in the slamfile then an empty string is returned but no warning
	is issued.
	"""
	args = profile.args(directive) or ['']
	if len(args) > 1: warn(directive + ' directive in ' + profile.get_path() + ' has multiple arguments, using ' + args[0])
	return args[0]


def tool_flags(ts, profile, tool_name):
	"""
	Gets the extra flags from the directives given by the name of the toolset ts and its base
	class toolsets. For example, for a C++ build, tool_name might be 'compiler' or 'linker'.

	This is a rather ugly hack and will be replaced by conditional areas of slamfiles in future.
	"""
	names = ts.names()
	flags = []
	for n in names:
		values = profile.args(n) or []
		for v in values:
			v = v.strip();
			if v.startswith(tool_name):
				v = v[len(tool_name):].strip()
				if v.startswith(':'):
					flags += v[1:].strip().split()

	return flags
